import copy

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Function


# ********************* 二值(+-1) ***********************
# activation
class BinaryActivation(Function):
    @staticmethod
    def forward(self, input):
        self.save_for_backward(input)
        output = torch.sign(input)
        output[output == 0] = 1
        # ******************** A —— 1、0 *********************
        # output = torch.clamp(output, min=0)
        return output

    @staticmethod
    def backward(self, grad_output):
        (input,) = self.saved_tensors
        # *******************ste*********************
        grad_input = grad_output.clone()
        # ****************saturate_ste***************
        grad_input[input.ge(1.0)] = 0
        grad_input[input.le(-1.0)] = 0
        """
        #******************soft_ste*****************
        size = input.size()
        zeros = torch.zeros(size).cuda()
        grad = torch.max(zeros, 1 - torch.abs(input))
        grad_input = grad_output * grad
        """
        return grad_input


# weight
class BinaryWeight(Function):
    @staticmethod
    def forward(self, input):
        output = torch.sign(input)
        output[output == 0] = 1
        return output

    @staticmethod
    def backward(self, grad_output):
        # *******************ste*********************
        grad_input = grad_output.clone()
        return grad_input


# ********************* 三值(+-1、0) ***********************
class Ternary(Function):
    @staticmethod
    def forward(self, input):
        # **************** channel级 - E(|W|) ****************
        E = torch.mean(torch.abs(input), (3, 2, 1), keepdim=True)
        # **************** 阈值 ****************
        threshold = E * 0.7
        # ************** W —— +-1、0 **************
        output = torch.sign(
            torch.add(
                torch.sign(torch.add(input, threshold)),
                torch.sign(torch.add(input, -threshold)),
            )
        )
        return output, threshold

    @staticmethod
    def backward(self, grad_output, grad_threshold):
        # *******************ste*********************
        grad_input = grad_output.clone()
        return grad_input


# ********************* A(特征)量化(二值) ***********************
class ActivationQuantizer(nn.Module):
    def __init__(self, A=2):
        super(ActivationQuantizer, self).__init__()
        self.A = A
        self.relu = nn.ReLU(inplace=True)

    def binary(self, input):
        output = BinaryActivation.apply(input)
        return output

    def forward(self, input):
        if self.A == 2:
            output = self.binary(input)
        else:
            output = self.relu(input)
        return output


# ********************* W(模型参数)量化(三/二值) ***********************
def meancenter_clamp_convparams(w):
    mean = w.data.mean(1, keepdim=True)
    w.data.sub_(mean)  # W中心化(C方向)
    w.data.clamp_(-1.0, 1.0)  # W截断
    return w


class WeightQuantizer(nn.Module):
    def __init__(self, W=2):
        super(WeightQuantizer, self).__init__()
        self.W = W

    def binary(self, input):
        output = BinaryWeight.apply(input)
        return output

    def ternary(self, input):
        output = Ternary.apply(input)
        return output

    def forward(self, input):
        if self.W == 2 or self.W == 3:
            # **************************************** W二值 *****************************************
            if self.W == 2:
                output = meancenter_clamp_convparams(input)  # W中心化+截断
                # **************** channel级 - E(|W|) ****************
                E = torch.mean(torch.abs(output), (3, 2, 1), keepdim=True)
                # **************** α(缩放因子) ****************
                alpha = E
                # ************** W —— +-1 **************
                output = self.binary(output)
                # ************** W * α **************
                output = output * alpha  # 若不需要α(缩放因子)，注释掉即可
            # **************************************** W三值 *****************************************
            elif self.W == 3:
                output_fp = input.clone()
                # ************** W —— +-1、0 **************
                output, threshold = self.ternary(input)  # threshold(阈值)
                # **************** α(缩放因子) ****************
                output_abs = torch.abs(output_fp)
                mask_le = output_abs.le(threshold)
                mask_gt = output_abs.gt(threshold)
                output_abs[mask_le] = 0
                output_abs_th = output_abs.clone()
                output_abs_th_sum = torch.sum(output_abs_th, (3, 2, 1), keepdim=True)
                mask_gt_sum = torch.sum(mask_gt, (3, 2, 1), keepdim=True).float()
                alpha = output_abs_th_sum / mask_gt_sum  # α(缩放因子)
                # *************** W * α ****************
                output = output * alpha  # 若不需要α(缩放因子)，注释掉即可
        else:
            output = input
        return output


class QuantConv2d(nn.Conv2d):
    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        stride=1,
        padding=0,
        dilation=1,
        groups=1,
        bias=True,
        padding_mode="zeros",
        W=2,
        quant_inference=False,
    ):
        super(QuantConv2d, self).__init__(
            in_channels,
            out_channels,
            kernel_size,
            stride,
            padding,
            dilation,
            groups,
            bias,
            padding_mode,
        )
        self.quant_inference = quant_inference
        self.weight_quantizer = WeightQuantizer(W=W)

    def forward(self, input):
        if not self.quant_inference:
            tnn_bin_weight = self.weight_quantizer(self.weight)
        else:
            tnn_bin_weight = self.weight
        output = F.conv2d(
            input,
            tnn_bin_weight,
            self.bias,
            self.stride,
            self.padding,
            self.dilation,
            self.groups,
        )
        return output


class QuantConvTranspose2d(nn.ConvTranspose2d):
    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        stride=1,
        padding=0,
        output_padding=0,
        dilation=1,
        groups=1,
        bias=True,
        padding_mode="zeros",
        W=2,
        quant_inference=False,
    ):
        super(QuantConvTranspose2d, self).__init__(
            in_channels,
            out_channels,
            kernel_size,
            stride,
            padding,
            output_padding,
            dilation,
            groups,
            bias,
            padding_mode,
        )
        self.quant_inference = quant_inference
        self.weight_quantizer = WeightQuantizer(W=W)

    def forward(self, input):
        if not self.quant_inference:
            tnn_bin_weight = self.weight_quantizer(self.weight)
        else:
            tnn_bin_weight = self.weight
        output = F.conv_transpose2d(
            input,
            tnn_bin_weight,
            self.bias,
            self.stride,
            self.padding,
            self.output_padding,
            self.groups,
            self.dilation,
        )
        return output


def add_quant_op(module, layer_counter, layer_num, A=2, W=2, quant_inference=False):
    for name, child in module.named_children():
        if isinstance(child, nn.Conv2d):
            layer_counter[0] += 1
            if layer_counter[0] > 1 and layer_counter[0] < layer_num:
                if child.bias is not None:
                    quant_conv = QuantConv2d(
                        child.in_channels,
                        child.out_channels,
                        child.kernel_size,
                        stride=child.stride,
                        padding=child.padding,
                        dilation=child.dilation,
                        groups=child.groups,
                        bias=True,
                        padding_mode=child.padding_mode,
                        W=W,
                        quant_inference=quant_inference,
                    )
                    quant_conv.bias.data = child.bias
                else:
                    quant_conv = QuantConv2d(
                        child.in_channels,
                        child.out_channels,
                        child.kernel_size,
                        stride=child.stride,
                        padding=child.padding,
                        dilation=child.dilation,
                        groups=child.groups,
                        bias=False,
                        padding_mode=child.padding_mode,
                        W=W,
                        quant_inference=quant_inference,
                    )
                quant_conv.weight.data = child.weight
                module._modules[name] = quant_conv
        elif isinstance(child, nn.ConvTranspose2d):
            layer_counter[0] += 1
            if layer_counter[0] > 1 and layer_counter[0] < layer_num:
                if child.bias is not None:
                    quant_conv_transpose = QuantConvTranspose2d(
                        child.in_channels,
                        child.out_channels,
                        child.kernel_size,
                        stride=child.stride,
                        padding=child.padding,
                        output_padding=child.output_padding,
                        dilation=child.dilation,
                        groups=child.groups,
                        bias=True,
                        padding_mode=child.padding_mode,
                        W=W,
                        quant_inference=quant_inference,
                    )
                    quant_conv_transpose.bias.data = child.bias
                else:
                    quant_conv_transpose = QuantConvTranspose2d(
                        child.in_channels,
                        child.out_channels,
                        child.kernel_size,
                        stride=child.stride,
                        padding=child.padding,
                        output_padding=child.output_padding,
                        dilation=child.dilation,
                        groups=child.groups,
                        bias=False,
                        padding_mode=child.padding_mode,
                        W=W,
                        quant_inference=quant_inference,
                    )
                quant_conv_transpose.weight.data = child.weight
                module._modules[name] = quant_conv_transpose
        elif isinstance(child, nn.ReLU):
            if layer_counter[0] > 0 and layer_counter[0] < layer_num:
                quant_relu = ActivationQuantizer(A=A)
                module._modules[name] = quant_relu
        else:
            add_quant_op(
                child,
                layer_counter,
                layer_num,
                A=A,
                W=W,
                quant_inference=quant_inference,
            )


def prepare(model, inplace=False, A=2, W=2, quant_inference=False):
    if not inplace:
        model = copy.deepcopy(model)
    layer_counter = [0]
    layer_num = 0
    for m in model.modules():
        if isinstance(m, nn.Conv2d):
            layer_num += 1
        elif isinstance(m, nn.ConvTranspose2d):
            layer_num += 1
    add_quant_op(
        model, layer_counter, layer_num, A=A, W=W, quant_inference=quant_inference
    )
    return model



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import sys

sys.path.append("../../..")
import os
import math
import argparse
import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim
import torchvision
import torchvision.transforms as transforms
from torch.autograd import Variable
from torch.nn import init
from models import nin_gc, nin

import quantize


def setup_seed(seed):
    torch.manual_seed(seed)
    # torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    np.random.seed(seed)
    torch.backends.cudnn.deterministic = True


def save_state(model, best_acc):
    print("==> Saving model ...")
    state = {
        "best_acc": best_acc,
        "state_dict": model.state_dict(),
    }
    state_copy = state["state_dict"].copy()
    for key in state_copy.keys():
        if "module" in key:
            state["state_dict"][key.replace("module.", "")] = state["state_dict"].pop(
                key
            )
    if args.model_type == 0:
        if args.prune_qat:
            torch.save(
                {"cfg": cfg, "best_acc": best_acc, "state_dict": state["state_dict"]},
                "models_save/nin.pth",
            )
        else:
            torch.save(state, "models_save/nin.pth")
    else:
        if args.prune_qat:
            torch.save(
                {"cfg": cfg, "best_acc": best_acc, "state_dict": state["state_dict"]},
                "models_save/nin_gc.pth",
            )
        else:
            torch.save(state, "models_save/nin_gc.pth")


def adjust_learning_rate(optimizer, epoch):
    update_list = [80, 130, 180, 230, 280]
    if epoch in update_list:
        for param_group in optimizer.param_groups:
            param_group["lr"] = param_group["lr"] * 0.1
    return


def train(epoch):
    model.train()

    for batch_idx, (data, target) in enumerate(trainloader):
        # 前向传播
        if not args.cpu:
            data, target = data.cuda(), target.cuda()
        data, target = Variable(data), Variable(target)
        output = model(data)
        loss = criterion(output, target)

        # 反向传播
        optimizer.zero_grad()
        loss.backward()  # 求梯度
        optimizer.step()  # 参数更新

        # 显示训练集loss(/100个batch)
        if batch_idx % 100 == 0:
            print(
                "Train Epoch: {} [{}/{} ({:.0f}%)]\tLoss: {:.6f}\tLR: {}".format(
                    epoch,
                    batch_idx * len(data),
                    len(trainloader.dataset),
                    100.0 * batch_idx / len(trainloader),
                    loss.data.item(),
                    optimizer.param_groups[0]["lr"],
                )
            )
    return


def test():
    global best_acc
    model.eval()
    test_loss = 0
    correct = 0

    for data, target in testloader:
        if not args.cpu:
            data, target = data.cuda(), target.cuda()
        data, target = Variable(data), Variable(target)
        # 前向传播
        output = model(data)
        test_loss += criterion(output, target).data.item()
        pred = output.data.max(1, keepdim=True)[1]
        correct += pred.eq(target.data.view_as(pred)).cpu().sum()
    # 测试准确率
    acc = 100.0 * float(correct) / len(testloader.dataset)

    # 最优准确率及model保存
    if acc > best_acc:
        best_acc = acc
        save_state(model, best_acc)
    average_test_loss = test_loss / (len(testloader.dataset) / args.eval_batch_size)

    # 显示测试集损失、准确率
    print(
        "\nTest set: Average loss: {:.4f}, Accuracy: {}/{} ({:.2f}%)".format(
            average_test_loss,
            correct,
            len(testloader.dataset),
            100.0 * float(correct) / len(testloader.dataset),
        )
    )

    # 显示测试集最优准确率
    print("Best Accuracy: {:.2f}%\n".format(best_acc))
    return


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    # cpu、gpu
    parser.add_argument(
        "--cpu", action="store_true", help="set if only CPU is available"
    )
    # gpu_id
    parser.add_argument("--gpu_id", action="store", default="", help="gpu_id")
    # dataset
    parser.add_argument(
        "--data", action="store", default="../../../data", help="dataset path"
    )
    # lr
    parser.add_argument(
        "--lr", action="store", default=0.01, help="the intial learning rate"
    )
    # weight_dacay
    parser.add_argument("--wd", action="store", default=0, help="nin_gc:0, nin:1e-5")
    # prune_qat
    parser.add_argument(
        "--prune_qat",
        default="",
        type=str,
        metavar="PATH",
        help="the path to the prune_qat model",
    )
    # refine
    parser.add_argument(
        "--refine",
        default="",
        type=str,
        metavar="PATH",
        help="the path to the float_refine model",
    )
    # resume
    parser.add_argument(
        "--resume",
        default="",
        type=str,
        metavar="PATH",
        help="the path to the resume model",
    )
    # batch_size、num_workers
    parser.add_argument("--train_batch_size", type=int, default=50)
    parser.add_argument("--eval_batch_size", type=int, default=256)
    parser.add_argument("--num_workers", type=int, default=2)
    # epochs
    parser.add_argument(
        "--start_epochs",
        type=int,
        default=1,
        metavar="N",
        help="number of epochs to train_start",
    )
    parser.add_argument(
        "--end_epochs",
        type=int,
        default=300,
        metavar="N",
        help="number of epochs to train_end",
    )
    # W/A — FP/三值/二值
    parser.add_argument("--W", type=int, default=2, help="Wb:2, Wt:3, Wfp:32")
    parser.add_argument("--A", type=int, default=2, help="Ab:2, Afp:32")
    # 模型结构选择
    parser.add_argument(
        "--model_type", type=int, default=1, help="model type:0-nin,1-nin_gc"
    )
    args = parser.parse_args()
    print("==> Options:", args)

    if args.gpu_id:
        os.environ["CUDA_VISIBLE_DEVICES"] = args.gpu_id

    setup_seed(1)  # 随机种子——训练结果可复现

    print("==> Preparing data..")
    # 数据增强
    # 训练集：随机裁剪 + 水平翻转 + 归一化
    transform_train = transforms.Compose(
        [
            transforms.RandomCrop(32, padding=4),
            transforms.RandomHorizontalFlip(),
            transforms.ToTensor(),
            transforms.Normalize((0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010)),
        ]
    )
    # 测试集：归一化
    transform_test = transforms.Compose(
        [
            transforms.ToTensor(),
            transforms.Normalize((0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010)),
        ]
    )

    # 数据加载
    trainset = torchvision.datasets.CIFAR10(
        root=args.data, train=True, download=True, transform=transform_train
    )
    trainloader = torch.utils.data.DataLoader(
        trainset,
        batch_size=args.train_batch_size,
        shuffle=True,
        num_workers=args.num_workers,
    )  # 训练集数据

    testset = torchvision.datasets.CIFAR10(
        root=args.data, train=False, download=True, transform=transform_test
    )
    testloader = torch.utils.data.DataLoader(
        testset,
        batch_size=args.eval_batch_size,
        shuffle=False,
        num_workers=args.num_workers,
    )  # 测试集数据

    # cifar10类别
    classes = (
        "plane",
        "car",
        "bird",
        "cat",
        "deer",
        "dog",
        "frog",
        "horse",
        "ship",
        "truck",
    )

    # model
    if args.prune_qat:
        print("******Prune Refine model******")
        # checkpoint = torch.load('../prune/models_save/nin_refine.pth')
        checkpoint = torch.load(args.prune_qat)
        cfg = checkpoint["cfg"]
        if args.model_type == 0:
            model = nin.Net(cfg=checkpoint["cfg"])
        else:
            model = nin_gc.Net(cfg=checkpoint["cfg"])
        model.load_state_dict(checkpoint["state_dict"])
        best_acc = 0
    elif args.refine:
        print("******Float Refine model******")
        # checkpoint = torch.load('models_save/nin.pth')
        state_dict = torch.load(args.refine)
        if args.model_type == 0:
            model = nin.Net()
        else:
            model = nin_gc.Net()
        model.load_state_dict(state_dict)
        best_acc = 0
    elif args.resume:
        print("******Reume model******")
        # checkpoint = torch.load('models_save/nin.pth')
        checkpoint = torch.load(args.resume)
        if args.model_type == 0:
            model = nin.Net()
        else:
            model = nin_gc.Net()
        model.load_state_dict(checkpoint["state_dict"])
        best_acc = checkpoint["best_acc"]
    else:
        print("******Initializing model******")
        if args.model_type == 0:
            model = nin.Net()
        else:
            model = nin_gc.Net()
        best_acc = 0
        for m in model.modules():
            if isinstance(m, nn.Conv2d):
                init.xavier_uniform_(m.weight)
                if m.bias is not None:
                    init.zeros_(m.bias)
            elif isinstance(m, nn.Linear):
                init.normal_(m.weight, 0, 0.01)
                if m.bias is not None:
                    init.zeros_(m.bias)
    print("***ori_model***\n", model)
    quantize.prepare(model, inplace=True, A=args.A, W=args.W)
    print("\n***quant_model***\n", model)

    # cpu、gpu
    if not args.cpu:
        model.cuda()
        model = torch.nn.DataParallel(
            model, device_ids=range(torch.cuda.device_count())
        )  # gpu并行训练

    # 超参数
    base_lr = float(args.lr)
    param_dict = dict(model.named_parameters())
    params = []
    for key, value in param_dict.items():
        params += [{"params": [value], "lr": base_lr, "weight_decay": args.wd}]

    # 损失函数
    criterion = nn.CrossEntropyLoss()
    # 优化器
    optimizer = optim.Adam(params, lr=base_lr, weight_decay=args.wd)

    # 训练模型
    for epoch in range(args.start_epochs, args.end_epochs):
        adjust_learning_rate(optimizer, epoch)
        train(epoch)
        test()



import sys

sys.path.append("../..")
import os
import argparse
import numpy as np
import torch
import torch.nn as nn
from torch.autograd import Variable
from torchvision import datasets, transforms
from models import nin


parser = argparse.ArgumentParser()
parser.add_argument("--data", action="store", default="../../data", help="dataset path")
parser.add_argument("--cpu", action="store_true", help="disables CUDA training")
# percent(剪枝率)
parser.add_argument("--percent", type=float, default=0.5, help="nin:0.5")
# 正常|规整剪枝标志
parser.add_argument(
    "--normal_regular",
    type=int,
    default=1,
    help="--normal_regular_flag (default: normal)",
)
# model层数
parser.add_argument("--layers", type=int, default=9, help="layers (default: 9)")
# 稀疏训练后的model
parser.add_argument(
    "--model",
    default="models_save/nin_preprune.pth",
    type=str,
    metavar="PATH",
    help="path to raw trained model (default: none)",
)
# 剪枝后保存的model
parser.add_argument(
    "--save",
    default="models_save/nin_prune.pth",
    type=str,
    metavar="PATH",
    help="path to save prune model (default: none)",
)
args = parser.parse_args()
base_number = args.normal_regular
layers = args.layers
print(args)

if base_number <= 0:
    print("\r\n!base_number is error!\r\n")
    base_number = 1

model = nin.Net()
if args.model:
    if os.path.isfile(args.model):
        print("=> loading checkpoint '{}'".format(args.model))
        model.load_state_dict(torch.load(args.model)["state_dict"])
    else:
        print("=> no checkpoint found at '{}'".format(args.resume))
print("旧模型: ", model)
total = 0
i = 0
for m in model.modules():
    if isinstance(m, nn.BatchNorm2d):
        if i < layers - 1:
            i += 1
            total += m.weight.data.shape[0]

# 确定剪枝的全局阈值
bn = torch.zeros(total)
index = 0
i = 0
for m in model.modules():
    if isinstance(m, nn.BatchNorm2d):
        if i < layers - 1:
            i += 1
            size = m.weight.data.shape[0]
            bn[index : (index + size)] = m.weight.data.abs().clone()
            index += size
y, j = torch.sort(bn)
thre_index = int(total * args.percent)
if thre_index == total:
    thre_index = total - 1
thre_0 = y[thre_index]

# ********************************预剪枝*********************************
pruned = 0
cfg_0 = []
cfg = []
cfg_mask = []
i = 0
for k, m in enumerate(model.modules()):
    if isinstance(m, nn.BatchNorm2d):
        if i < layers - 1:
            i += 1

            weight_copy = m.weight.data.clone()
            mask = weight_copy.abs().gt(thre_0).float()
            remain_channels = torch.sum(mask)

            if remain_channels == 0:
                print("\r\n!please turn down the prune_ratio!\r\n")
                remain_channels = 1
                mask[int(torch.argmax(weight_copy))] = 1

            # ******************规整剪枝******************
            v = 0
            n = 1
            if remain_channels % base_number != 0:
                if remain_channels > base_number:
                    while v < remain_channels:
                        n += 1
                        v = base_number * n
                    if remain_channels - (v - base_number) < v - remain_channels:
                        remain_channels = v - base_number
                    else:
                        remain_channels = v
                    if remain_channels > m.weight.data.size()[0]:
                        remain_channels = m.weight.data.size()[0]
                    remain_channels = torch.tensor(remain_channels)

                    y, j = torch.sort(weight_copy.abs())
                    thre_1 = y[-remain_channels]
                    mask = weight_copy.abs().ge(thre_1).float()
            pruned = pruned + mask.shape[0] - torch.sum(mask)
            m.weight.data.mul_(mask)
            m.bias.data.mul_(mask)
            cfg_0.append(mask.shape[0])
            cfg.append(int(remain_channels))
            cfg_mask.append(mask.clone())
            print(
                "layer_index: {:d} \t total_channel: {:d} \t remaining_channel: {:d} \t pruned_ratio: {:f}".format(
                    k,
                    mask.shape[0],
                    int(torch.sum(mask)),
                    (mask.shape[0] - torch.sum(mask)) / mask.shape[0],
                )
            )
pruned_ratio = float(pruned / total)
print("\r\n!预剪枝完成!")
print("total_pruned_ratio: ", pruned_ratio)

# ********************************预剪枝后model测试*********************************
def test():
    test_loader = torch.utils.data.DataLoader(
        datasets.CIFAR10(
            root=args.data,
            train=False,
            transform=transforms.Compose(
                [
                    transforms.ToTensor(),
                    transforms.Normalize(
                        (0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010)
                    ),
                ]
            ),
        ),
        batch_size=64,
        shuffle=False,
        num_workers=1,
    )
    model.eval()
    correct = 0

    for data, target in test_loader:
        if not args.cpu:
            data, target = data.cuda(), target.cuda()
        data, target = Variable(data), Variable(target)
        output = model(data)
        pred = output.data.max(1, keepdim=True)[1]
        correct += pred.eq(target.data.view_as(pred)).cpu().sum()
    acc = 100.0 * float(correct) / len(test_loader.dataset)
    print("Accuracy: {:.2f}%\n".format(acc))
    return


print("************预剪枝模型测试************")
if not args.cpu:
    model.cuda()
test()

# ********************************剪枝*********************************
newmodel = nin.Net(cfg)
if not args.cpu:
    newmodel.cuda()
layer_id_in_cfg = 0
start_mask = torch.ones(3)
end_mask = cfg_mask[layer_id_in_cfg]
i = 0
for [m0, m1] in zip(model.modules(), newmodel.modules()):
    if isinstance(m0, nn.BatchNorm2d):
        if i < layers - 1:
            i += 1
            idx1 = np.squeeze(np.argwhere(np.asarray(end_mask.cpu().numpy())))
            if idx1.size == 1:
                idx1 = np.resize(idx1, (1,))
            m1.weight.data = m0.weight.data[idx1].clone()
            m1.bias.data = m0.bias.data[idx1].clone()
            m1.running_mean = m0.running_mean[idx1].clone()
            m1.running_var = m0.running_var[idx1].clone()
            layer_id_in_cfg += 1
            start_mask = end_mask.clone()
            if layer_id_in_cfg < len(cfg_mask):
                end_mask = cfg_mask[layer_id_in_cfg]
        else:
            m1.weight.data = m0.weight.data.clone()
            m1.bias.data = m0.bias.data.clone()
            m1.running_mean = m0.running_mean.clone()
            m1.running_var = m0.running_var.clone()
    elif isinstance(m0, nn.Conv2d):
        if i < layers - 1:
            idx0 = np.squeeze(np.argwhere(np.asarray(start_mask.cpu().numpy())))
            idx1 = np.squeeze(np.argwhere(np.asarray(end_mask.cpu().numpy())))
            if idx0.size == 1:
                idx0 = np.resize(idx0, (1,))
            if idx1.size == 1:
                idx1 = np.resize(idx1, (1,))
            w = m0.weight.data[:, idx0, :, :].clone()
            m1.weight.data = w[idx1, :, :, :].clone()
            m1.bias.data = m0.bias.data[idx1].clone()
        else:
            idx0 = np.squeeze(np.argwhere(np.asarray(start_mask.cpu().numpy())))
            if idx0.size == 1:
                idx0 = np.resize(idx0, (1,))
            m1.weight.data = m0.weight.data[:, idx0, :, :].clone()
            m1.bias.data = m0.bias.data.clone()
    elif isinstance(m0, nn.Linear):
        idx0 = np.squeeze(np.argwhere(np.asarray(start_mask.cpu().numpy())))
        if idx0.size == 1:
            idx0 = np.resize(idx0, (1,))
        m1.weight.data = m0.weight.data[:, idx0].clone()

# ******************************剪枝后model测试*********************************
print("新模型: ", newmodel)
print("**********剪枝后新模型测试*********")
model = newmodel
test()

# ******************************剪枝后model保存*********************************
print("**********剪枝后新模型保存*********")
torch.save({"cfg": cfg, "state_dict": newmodel.state_dict()}, args.save)
print("**********保存成功*********\r\n")

# *****************************剪枝前后model对比********************************
print("************旧模型结构************")
print(cfg_0)
print("************新模型结构************")
print(cfg, "\r\n")



__version__ = "1.12.0"

from micronet.base_module.op import *


def quant_test_manual():
    import torch.nn as nn
    import torch.nn.functional as F

    # ``quantize`` is quant_module, ``QuantConv2d``, ``QuantLinear``, ``QuantMaxPool2d``, ``QuantReLU`` are quant_op
    from micronet.compression.quantization.wbwtab.quantize import (
        QuantConv2d as quant_conv_wbwtab,
    )
    from micronet.compression.quantization.wbwtab.quantize import (
        ActivationQuantizer as quant_relu_wbwtab,
    )
    from micronet.compression.quantization.wqaq.dorefa.quantize import (
        QuantConv2d as quant_conv_dorefa,
    )
    from micronet.compression.quantization.wqaq.dorefa.quantize import (
        QuantLinear as quant_linear_dorefa,
    )
    from micronet.compression.quantization.wqaq.iao.quantize import (
        QuantConv2d as quant_conv_iao,
    )
    from micronet.compression.quantization.wqaq.iao.quantize import (
        QuantLinear as quant_linear_iao,
    )
    from micronet.compression.quantization.wqaq.iao.quantize import (
        QuantMaxPool2d as quant_max_pool_iao,
    )
    from micronet.compression.quantization.wqaq.iao.quantize import (
        QuantReLU as quant_relu_iao,
    )

    class LeNet(nn.Module):
        def __init__(self):
            super(LeNet, self).__init__()
            self.conv1 = nn.Conv2d(1, 10, kernel_size=5)
            self.conv2 = nn.Conv2d(10, 20, kernel_size=5)
            self.fc1 = nn.Linear(320, 50)
            self.fc2 = nn.Linear(50, 10)
            self.max_pool = nn.MaxPool2d(kernel_size=2)
            self.relu = nn.ReLU(inplace=True)

        def forward(self, x):
            x = self.relu(self.max_pool(self.conv1(x)))
            x = self.relu(self.max_pool(self.conv2(x)))
            x = x.view(-1, 320)
            x = self.relu(self.fc1(x))
            x = F.dropout(x, training=self.training)
            x = self.fc2(x)
            return F.log_softmax(x, dim=1)

    class QuantLeNetWbWtAb(nn.Module):
        def __init__(self):
            super(QuantLeNetWbWtAb, self).__init__()
            self.conv1 = quant_conv_wbwtab(1, 10, kernel_size=5)
            self.conv2 = quant_conv_wbwtab(10, 20, kernel_size=5)
            self.fc1 = nn.Linear(320, 50)
            self.fc2 = nn.Linear(50, 10)
            self.max_pool = nn.MaxPool2d(kernel_size=2)
            self.relu = quant_relu_wbwtab()

        def forward(self, x):
            x = self.relu(self.max_pool(self.conv1(x)))
            x = self.relu(self.max_pool(self.conv2(x)))
            x = x.view(-1, 320)
            x = self.relu(self.fc1(x))
            x = F.dropout(x, training=self.training)
            x = self.fc2(x)
            return F.log_softmax(x, dim=1)

    class QuantLeNetDoReFa(nn.Module):
        def __init__(self):
            super(QuantLeNetDoReFa, self).__init__()
            self.conv1 = quant_conv_dorefa(1, 10, kernel_size=5)
            self.conv2 = quant_conv_dorefa(10, 20, kernel_size=5)
            self.fc1 = quant_linear_dorefa(320, 50)
            self.fc2 = quant_linear_dorefa(50, 10)
            self.max_pool = nn.MaxPool2d(kernel_size=2)
            self.relu = nn.ReLU(inplace=True)

        def forward(self, x):
            x = self.relu(self.max_pool(self.conv1(x)))
            x = self.relu(self.max_pool(self.conv2(x)))
            x = x.view(-1, 320)
            x = self.relu(self.fc1(x))
            x = F.dropout(x, training=self.training)
            x = self.fc2(x)
            return F.log_softmax(x, dim=1)

    class QuantLeNetIAO(nn.Module):
        def __init__(self):
            super(QuantLeNetIAO, self).__init__()
            self.conv1 = quant_conv_iao(1, 10, kernel_size=5)
            self.conv2 = quant_conv_iao(10, 20, kernel_size=5)
            self.fc1 = quant_linear_iao(320, 50)
            self.fc2 = quant_linear_iao(50, 10)
            self.max_pool = quant_max_pool_iao(kernel_size=2)
            self.relu = nn.ReLU(inplace=True)

        def forward(self, x):
            x = self.relu(self.max_pool(self.conv1(x)))
            x = self.relu(self.max_pool(self.conv2(x)))
            x = x.view(-1, 320)
            x = self.relu(self.fc1(x))
            x = F.dropout(x, training=self.training)
            x = self.fc2(x)
            return F.log_softmax(x, dim=1)

    lenet = LeNet()
    quant_lenet_wbwtab = QuantLeNetWbWtAb()
    quant_lenet_dorefa = QuantLeNetDoReFa()
    quant_lenet_iao = QuantLeNetIAO()

    print("***ori_model***\n", lenet)
    print("\n***quant_model_wbwtab***\n", quant_lenet_wbwtab)
    print("\n***quant_model_dorefa***\n", quant_lenet_dorefa)
    print("\n***quant_model_iao***\n", quant_lenet_iao)

    print("\nquant_model is ready")
    print("micronet is ready")


def quant_test_auto():
    import torch.nn as nn
    import torch.nn.functional as F

    import micronet.compression.quantization.wqaq.dorefa.quantize as quant_dorefa
    import micronet.compression.quantization.wqaq.iao.quantize as quant_iao

    class LeNet(nn.Module):
        def __init__(self):
            super(LeNet, self).__init__()
            self.conv1 = nn.Conv2d(1, 10, kernel_size=5)
            self.conv2 = nn.Conv2d(10, 20, kernel_size=5)
            self.fc1 = nn.Linear(320, 50)
            self.fc2 = nn.Linear(50, 10)
            self.max_pool = nn.MaxPool2d(kernel_size=2)
            self.relu = nn.ReLU(inplace=True)

        def forward(self, x):
            x = self.relu(self.max_pool(self.conv1(x)))
            x = self.relu(self.max_pool(self.conv2(x)))
            x = x.view(-1, 320)
            x = self.relu(self.fc1(x))
            x = F.dropout(x, training=self.training)
            x = self.fc2(x)
            return F.log_softmax(x, dim=1)

    lenet = LeNet()
    quant_lenet_dorefa = quant_dorefa.prepare(lenet, inplace=False, a_bits=8, w_bits=8)
    quant_lenet_iao = quant_iao.prepare(
        lenet,
        inplace=False,
        a_bits=8,
        w_bits=8,
        q_type=0,
        q_level=0,
        weight_observer=0,
        bn_fuse=False,
        bn_fuse_calib=False,
        pretrained_model=False,
        qaft=False,
        ptq=False,
        percentile=0.9999,
    )

    print("***ori_model***\n", lenet)
    print("\n***quant_model_dorefa***\n", quant_lenet_dorefa)
    print("\n***quant_model_iao***\n", quant_lenet_iao)

    print("\nquant_model is ready")
    print("micronet is ready")



import copy

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import distributed
from torch.nn import init
from torch.nn.parameter import Parameter
from torch.autograd import Function

from micronet.base_module.op import *


# ********************* observers(统计min/max) *********************
class ObserverBase(nn.Module):
    def __init__(self, q_level):
        super(ObserverBase, self).__init__()
        self.q_level = q_level

    def update_range(self, min_val, max_val):
        raise NotImplementedError

    @torch.no_grad()
    def forward(self, input):
        if self.q_level == "L":  # layer级(activation/weight)
            min_val = torch.min(input)
            max_val = torch.max(input)
        elif self.q_level == "C":  # channel级(conv_weight)
            input = torch.flatten(input, start_dim=1)
            min_val = torch.min(input, 1)[0]
            max_val = torch.max(input, 1)[0]
        elif self.q_level == "FC":  # channel级(fc_weight)
            min_val = torch.min(input, 1, keepdim=True)[0]
            max_val = torch.max(input, 1, keepdim=True)[0]

        self.update_range(min_val, max_val)


class MinMaxObserver(ObserverBase):
    def __init__(self, q_level, out_channels):
        super(MinMaxObserver, self).__init__(q_level)
        self.num_flag = 0
        self.out_channels = out_channels
        if self.q_level == "L":
            self.register_buffer("min_val", torch.zeros((1), dtype=torch.float32))
            self.register_buffer("max_val", torch.zeros((1), dtype=torch.float32))
        elif self.q_level == "C":
            self.register_buffer(
                "min_val", torch.zeros((out_channels, 1, 1, 1), dtype=torch.float32)
            )
            self.register_buffer(
                "max_val", torch.zeros((out_channels, 1, 1, 1), dtype=torch.float32)
            )
        elif self.q_level == "FC":
            self.register_buffer(
                "min_val", torch.zeros((out_channels, 1), dtype=torch.float32)
            )
            self.register_buffer(
                "max_val", torch.zeros((out_channels, 1), dtype=torch.float32)
            )

    def update_range(self, min_val_cur, max_val_cur):
        if self.q_level == "C":
            min_val_cur.resize_(self.min_val.shape)
            max_val_cur.resize_(self.max_val.shape)
        if self.num_flag == 0:
            self.num_flag += 1
            min_val = min_val_cur
            max_val = max_val_cur
        else:
            min_val = torch.min(min_val_cur, self.min_val)
            max_val = torch.max(max_val_cur, self.max_val)
        self.min_val.copy_(min_val)
        self.max_val.copy_(max_val)


class MovingAverageMinMaxObserver(ObserverBase):
    def __init__(self, q_level, out_channels, momentum=0.1):
        super(MovingAverageMinMaxObserver, self).__init__(q_level)
        self.momentum = momentum
        self.num_flag = 0
        self.out_channels = out_channels
        if self.q_level == "L":
            self.register_buffer("min_val", torch.zeros((1), dtype=torch.float32))
            self.register_buffer("max_val", torch.zeros((1), dtype=torch.float32))
        elif self.q_level == "C":
            self.register_buffer(
                "min_val", torch.zeros((out_channels, 1, 1, 1), dtype=torch.float32)
            )
            self.register_buffer(
                "max_val", torch.zeros((out_channels, 1, 1, 1), dtype=torch.float32)
            )
        elif self.q_level == "FC":
            self.register_buffer(
                "min_val", torch.zeros((out_channels, 1), dtype=torch.float32)
            )
            self.register_buffer(
                "max_val", torch.zeros((out_channels, 1), dtype=torch.float32)
            )

    def update_range(self, min_val_cur, max_val_cur):
        if self.q_level == "C":
            min_val_cur.resize_(self.min_val.shape)
            max_val_cur.resize_(self.max_val.shape)
        if self.num_flag == 0:
            self.num_flag += 1
            min_val = min_val_cur
            max_val = max_val_cur
        else:
            min_val = (1 - self.momentum) * self.min_val + self.momentum * min_val_cur
            max_val = (1 - self.momentum) * self.max_val + self.momentum * max_val_cur
        self.min_val.copy_(min_val)
        self.max_val.copy_(max_val)


class HistogramObserver(nn.Module):
    def __init__(self, q_level, momentum=0.1, percentile=0.9999):
        super(HistogramObserver, self).__init__()
        self.q_level = q_level
        self.momentum = momentum
        self.percentile = percentile
        self.num_flag = 0
        self.register_buffer("min_val", torch.zeros((1), dtype=torch.float32))
        self.register_buffer("max_val", torch.zeros((1), dtype=torch.float32))

    @torch.no_grad()
    def forward(self, input):
        # MovingAveragePercentileCalibrator
        # PercentileCalibrator
        max_val_cur = torch.kthvalue(
            input.abs().view(-1), int(self.percentile * input.view(-1).size(0)), dim=0
        )[0]
        # MovingAverage
        if self.num_flag == 0:
            self.num_flag += 1
            max_val = max_val_cur
        else:
            max_val = (1 - self.momentum) * self.max_val + self.momentum * max_val_cur
        self.max_val.copy_(max_val)


# ********************* quantizers（量化器，量化） *********************
# 取整(饱和/截断ste)
class Round(Function):
    @staticmethod
    def forward(self, input, observer_min_val, observer_max_val, q_type):
        # 对称
        if q_type == 0:
            max_val = torch.max(
                torch.abs(observer_min_val), torch.abs(observer_max_val)
            )
            min_val = -max_val
        # 非对称
        else:
            max_val = observer_max_val
            min_val = observer_min_val
        self.save_for_backward(input, min_val, max_val)
        sign = torch.sign(input)
        output = sign * torch.floor(torch.abs(input) + 0.5)
        return output

    @staticmethod
    def backward(self, grad_output):
        input, min_val, max_val = self.saved_tensors
        grad_input = grad_output.clone()
        grad_input[input.gt(max_val)] = 0
        grad_input[input.lt(min_val)] = 0
        return grad_input, None, None, None


class Quantizer(nn.Module):
    def __init__(self, bits, observer, activation_weight_flag, qaft=False, union=False):
        super(Quantizer, self).__init__()
        self.bits = bits
        self.observer = observer
        self.activation_weight_flag = activation_weight_flag
        self.qaft = qaft
        self.union = union
        self.q_type = 0
        # scale/zero_point/eps
        if self.observer.q_level == "L":
            self.register_buffer("scale", torch.ones((1), dtype=torch.float32))
            self.register_buffer("zero_point", torch.zeros((1), dtype=torch.float32))
        elif self.observer.q_level == "C":
            self.register_buffer(
                "scale",
                torch.ones((self.observer.out_channels, 1, 1, 1), dtype=torch.float32),
            )
            self.register_buffer(
                "zero_point",
                torch.zeros((self.observer.out_channels, 1, 1, 1), dtype=torch.float32),
            )
        elif self.observer.q_level == "FC":
            self.register_buffer(
                "scale",
                torch.ones((self.observer.out_channels, 1), dtype=torch.float32),
            )
            self.register_buffer(
                "zero_point",
                torch.zeros((self.observer.out_channels, 1), dtype=torch.float32),
            )
        self.register_buffer(
            "eps", torch.tensor((torch.finfo(torch.float32).eps), dtype=torch.float32)
        )

    def update_qparams(self):
        raise NotImplementedError

    # 取整(ste)
    def round(self, input, observer_min_val, observer_max_val, q_type):
        output = Round.apply(input, observer_min_val, observer_max_val, q_type)
        return output

    def forward(self, input):
        if self.bits == 32:
            output = input
        elif self.bits == 1:
            print("！Binary quantization is not supported ！")
            assert self.bits != 1
        else:
            if not self.qaft:
                # qat, update quant_para
                if self.training:
                    if not self.union:
                        self.observer(input)  # update observer_min and observer_max
                    self.update_qparams()  # update scale and zero_point
            output = (
                torch.clamp(
                    self.round(
                        input / self.scale.clone() - self.zero_point,
                        self.observer.min_val / self.scale - self.zero_point,
                        self.observer.max_val / self.scale - self.zero_point,
                        self.q_type,
                    ),
                    self.quant_min_val,
                    self.quant_max_val,
                )
                + self.zero_point
            ) * self.scale.clone()
        return output


class SignedQuantizer(Quantizer):
    def __init__(self, *args, **kwargs):
        super(SignedQuantizer, self).__init__(*args, **kwargs)
        if self.activation_weight_flag == 0:  # weight
            self.register_buffer(
                "quant_min_val",
                torch.tensor((-((1 << (self.bits - 1)) - 1)), dtype=torch.float32),
            )
            self.register_buffer(
                "quant_max_val",
                torch.tensor(((1 << (self.bits - 1)) - 1), dtype=torch.float32),
            )
        elif self.activation_weight_flag == 1:  # activation
            self.register_buffer(
                "quant_min_val",
                torch.tensor((-(1 << (self.bits - 1))), dtype=torch.float32),
            )
            self.register_buffer(
                "quant_max_val",
                torch.tensor(((1 << (self.bits - 1)) - 1), dtype=torch.float32),
            )
        else:
            print("activation_weight_flag error")


class UnsignedQuantizer(Quantizer):
    def __init__(self, *args, **kwargs):
        super(UnsignedQuantizer, self).__init__(*args, **kwargs)
        if self.activation_weight_flag == 0:  # weight
            self.register_buffer(
                "quant_min_val", torch.tensor((0), dtype=torch.float32)
            )
            self.register_buffer(
                "quant_max_val",
                torch.tensor(((1 << self.bits) - 2), dtype=torch.float32),
            )
        elif self.activation_weight_flag == 1:  # activation
            self.register_buffer(
                "quant_min_val", torch.tensor((0), dtype=torch.float32)
            )
            self.register_buffer(
                "quant_max_val",
                torch.tensor(((1 << self.bits) - 1), dtype=torch.float32),
            )
        else:
            print("activation_weight_flag error")


# 对称量化
class SymmetricQuantizer(SignedQuantizer):
    def update_qparams(self):
        self.q_type = 0
        quant_range = (
            float(self.quant_max_val - self.quant_min_val) / 2
        )  # quantized_range
        float_range = torch.max(
            torch.abs(self.observer.min_val), torch.abs(self.observer.max_val)
        )  # float_range
        scale = float_range / quant_range  # scale
        scale = torch.max(scale, self.eps)  # processing for very small scale
        zero_point = torch.zeros_like(scale)  # zero_point
        self.scale.copy_(scale)
        self.zero_point.copy_(zero_point)


# 非对称量化
class AsymmetricQuantizer(UnsignedQuantizer):
    def update_qparams(self):
        self.q_type = 1
        quant_range = float(self.quant_max_val - self.quant_min_val)  # quantized_range
        float_range = self.observer.max_val - self.observer.min_val  # float_range
        scale = float_range / quant_range  # scale
        scale = torch.max(scale, self.eps)  # processing for very small scale
        sign = torch.sign(self.observer.min_val)
        zero_point = sign * torch.floor(
            torch.abs(self.observer.min_val / scale) + 0.5
        )  # zero_point
        self.scale.copy_(scale)
        self.zero_point.copy_(zero_point)


# ********************* 量化卷积（同时量化A/W，并做卷积） *********************
class QuantConv2d(nn.Conv2d):
    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        stride=1,
        padding=0,
        dilation=1,
        groups=1,
        bias=True,
        padding_mode="zeros",
        a_bits=8,
        w_bits=8,
        q_type=0,
        q_level=0,
        weight_observer=0,
        quant_inference=False,
        qaft=False,
        ptq=False,
        percentile=0.9999,
    ):
        super(QuantConv2d, self).__init__(
            in_channels,
            out_channels,
            kernel_size,
            stride,
            padding,
            dilation,
            groups,
            bias,
            padding_mode,
        )
        self.quant_inference = quant_inference
        if not ptq:
            if q_type == 0:
                self.activation_quantizer = SymmetricQuantizer(
                    bits=a_bits,
                    observer=MovingAverageMinMaxObserver(
                        q_level="L", out_channels=None
                    ),
                    activation_weight_flag=1,
                    qaft=qaft,
                )
                if weight_observer == 0:
                    if q_level == 0:
                        self.weight_quantizer = SymmetricQuantizer(
                            bits=w_bits,
                            observer=MinMaxObserver(
                                q_level="C", out_channels=out_channels
                            ),
                            activation_weight_flag=0,
                            qaft=qaft,
                        )
                    else:
                        self.weight_quantizer = SymmetricQuantizer(
                            bits=w_bits,
                            observer=MinMaxObserver(q_level="L", out_channels=None),
                            activation_weight_flag=0,
                            qaft=qaft,
                        )
                else:
                    if q_level == 0:
                        self.weight_quantizer = SymmetricQuantizer(
                            bits=w_bits,
                            observer=MovingAverageMinMaxObserver(
                                q_level="C", out_channels=out_channels
                            ),
                            activation_weight_flag=0,
                            qaft=qaft,
                        )
                    else:
                        self.weight_quantizer = SymmetricQuantizer(
                            bits=w_bits,
                            observer=MovingAverageMinMaxObserver(
                                q_level="L", out_channels=None
                            ),
                            activation_weight_flag=0,
                            qaft=qaft,
                        )
            else:
                self.activation_quantizer = AsymmetricQuantizer(
                    bits=a_bits,
                    observer=MovingAverageMinMaxObserver(
                        q_level="L", out_channels=None
                    ),
                    activation_weight_flag=1,
                    qaft=qaft,
                )
                if weight_observer == 0:
                    if q_level == 0:
                        self.weight_quantizer = AsymmetricQuantizer(
                            bits=w_bits,
                            observer=MinMaxObserver(
                                q_level="C", out_channels=out_channels
                            ),
                            activation_weight_flag=0,
                            qaft=qaft,
                        )
                    else:
                        self.weight_quantizer = AsymmetricQuantizer(
                            bits=w_bits,
                            observer=MinMaxObserver(q_level="L", out_channels=None),
                            activation_weight_flag=0,
                            qaft=qaft,
                        )
                else:
                    if q_level == 0:
                        self.weight_quantizer = AsymmetricQuantizer(
                            bits=w_bits,
                            observer=MovingAverageMinMaxObserver(
                                q_level="C", out_channels=out_channels
                            ),
                            activation_weight_flag=0,
                            qaft=qaft,
                        )
                    else:
                        self.weight_quantizer = AsymmetricQuantizer(
                            bits=w_bits,
                            observer=MovingAverageMinMaxObserver(
                                q_level="L", out_channels=None
                            ),
                            activation_weight_flag=0,
                            qaft=qaft,
                        )
        else:
            self.activation_quantizer = SymmetricQuantizer(
                bits=a_bits,
                observer=HistogramObserver(q_level="L", percentile=percentile),
                activation_weight_flag=1,
                qaft=qaft,
            )
            if weight_observer == 0:
                if q_level == 0:
                    self.weight_quantizer = SymmetricQuantizer(
                        bits=w_bits,
                        observer=MinMaxObserver(q_level="C", out_channels=out_channels),
                        activation_weight_flag=0,
                        qaft=qaft,
                    )
                else:
                    self.weight_quantizer = SymmetricQuantizer(
                        bits=w_bits,
                        observer=MinMaxObserver(q_level="L", out_channels=None),
                        activation_weight_flag=0,
                        qaft=qaft,
                    )
            else:
                if q_level == 0:
                    self.weight_quantizer = SymmetricQuantizer(
                        bits=w_bits,
                        observer=MovingAverageMinMaxObserver(
                            q_level="C", out_channels=out_channels
                        ),
                        activation_weight_flag=0,
                        qaft=qaft,
                    )
                else:
                    self.weight_quantizer = SymmetricQuantizer(
                        bits=w_bits,
                        observer=MovingAverageMinMaxObserver(
                            q_level="L", out_channels=None
                        ),
                        activation_weight_flag=0,
                        qaft=qaft,
                    )

    def forward(self, input):
        quant_input = self.activation_quantizer(input)
        if not self.quant_inference:
            quant_weight = self.weight_quantizer(self.weight)
        else:
            quant_weight = self.weight
        output = F.conv2d(
            quant_input,
            quant_weight,
            self.bias,
            self.stride,
            self.padding,
            self.dilation,
            self.groups,
        )
        return output


class QuantConvTranspose2d(nn.ConvTranspose2d):
    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        stride=1,
        padding=0,
        output_padding=0,
        groups=1,
        bias=True,
        dilation=1,
        padding_mode="zeros",
        a_bits=8,
        w_bits=8,
        q_type=0,
        weight_observer=0,
        quant_inference=False,
        qaft=False,
        ptq=False,
        percentile=0.9999,
    ):
        super(QuantConvTranspose2d, self).__init__(
            in_channels,
            out_channels,
            kernel_size,
            stride,
            padding,
            output_padding,
            groups,
            bias,
            dilation,
            padding_mode,
        )
        self.quant_inference = quant_inference
        if not ptq:
            if q_type == 0:
                self.activation_quantizer = SymmetricQuantizer(
                    bits=a_bits,
                    observer=MovingAverageMinMaxObserver(
                        q_level="L", out_channels=None
                    ),
                    activation_weight_flag=1,
                    qaft=qaft,
                )
                if weight_observer == 0:
                    self.weight_quantizer = SymmetricQuantizer(
                        bits=w_bits,
                        observer=MinMaxObserver(q_level="L", out_channels=None),
                        activation_weight_flag=0,
                        qaft=qaft,
                    )
                else:
                    self.weight_quantizer = SymmetricQuantizer(
                        bits=w_bits,
                        observer=MovingAverageMinMaxObserver(
                            q_level="L", out_channels=None
                        ),
                        activation_weight_flag=0,
                        qaft=qaft,
                    )
            else:
                self.activation_quantizer = AsymmetricQuantizer(
                    bits=a_bits,
                    observer=MovingAverageMinMaxObserver(
                        q_level="L", out_channels=None
                    ),
                    activation_weight_flag=1,
                    qaft=qaft,
                )
                if weight_observer == 0:
                    self.weight_quantizer = AsymmetricQuantizer(
                        bits=w_bits,
                        observer=MinMaxObserver(q_level="L", out_channels=None),
                        activation_weight_flag=0,
                        qaft=qaft,
                    )
                else:
                    self.weight_quantizer = AsymmetricQuantizer(
                        bits=w_bits,
                        observer=MovingAverageMinMaxObserver(
                            q_level="L", out_channels=None
                        ),
                        activation_weight_flag=0,
                        qaft=qaft,
                    )
        else:
            self.activation_quantizer = SymmetricQuantizer(
                bits=a_bits,
                observer=HistogramObserver(q_level="L", percentile=percentile),
                activation_weight_flag=1,
                qaft=qaft,
            )
            if weight_observer == 0:
                self.weight_quantizer = SymmetricQuantizer(
                    bits=w_bits,
                    observer=MinMaxObserver(q_level="L", out_channels=None),
                    activation_weight_flag=0,
                    qaft=qaft,
                )
            else:
                self.weight_quantizer = SymmetricQuantizer(
                    bits=w_bits,
                    observer=MovingAverageMinMaxObserver(
                        q_level="L", out_channels=None
                    ),
                    activation_weight_flag=0,
                    qaft=qaft,
                )

    def forward(self, input):
        quant_input = self.activation_quantizer(input)
        if not self.quant_inference:
            quant_weight = self.weight_quantizer(self.weight)
        else:
            quant_weight = self.weight
        output = F.conv_transpose2d(
            quant_input,
            quant_weight,
            self.bias,
            self.stride,
            self.padding,
            self.output_padding,
            self.groups,
            self.dilation,
        )
        return output


def reshape_to_activation(input):
    return input.reshape(1, -1, 1, 1)


def reshape_to_weight(input):
    return input.reshape(-1, 1, 1, 1)


def reshape_to_bias(input):
    return input.reshape(-1)


# ********************* bn融合_量化卷积（bn融合后，同时量化A/W，并做卷积） *********************
class QuantBNFuseConv2d(QuantConv2d):
    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        stride=1,
        padding=0,
        dilation=1,
        groups=1,
        bias=False,
        padding_mode="zeros",
        eps=1e-5,
        momentum=0.1,
        a_bits=8,
        w_bits=8,
        q_type=0,
        q_level=0,
        weight_observer=0,
        pretrained_model=False,
        qaft=False,
        ptq=False,
        percentile=0.9999,
        bn_fuse_calib=False,
    ):
        super(QuantBNFuseConv2d, self).__init__(
            in_channels,
            out_channels,
            kernel_size,
            stride,
            padding,
            dilation,
            groups,
            bias,
            padding_mode,
        )
        self.num_flag = 0
        self.pretrained_model = pretrained_model
        self.qaft = qaft
        self.bn_fuse_calib = bn_fuse_calib
        self.eps = eps
        self.momentum = momentum
        self.gamma = Parameter(torch.Tensor(out_channels))
        self.beta = Parameter(torch.Tensor(out_channels))
        self.register_buffer(
            "running_mean", torch.zeros((out_channels), dtype=torch.float32)
        )
        self.register_buffer(
            "running_var", torch.ones((out_channels), dtype=torch.float32)
        )
        init.uniform_(self.gamma)
        init.zeros_(self.beta)
        if not ptq:
            if q_type == 0:
                self.activation_quantizer = SymmetricQuantizer(
                    bits=a_bits,
                    observer=MovingAverageMinMaxObserver(
                        q_level="L", out_channels=None
                    ),
                    activation_weight_flag=1,
                    qaft=qaft,
                )
                if weight_observer == 0:
                    if q_level == 0:
                        self.weight_quantizer = SymmetricQuantizer(
                            bits=w_bits,
                            observer=MinMaxObserver(
                                q_level="C", out_channels=out_channels
                            ),
                            activation_weight_flag=0,
                            qaft=qaft,
                        )
                    else:
                        self.weight_quantizer = SymmetricQuantizer(
                            bits=w_bits,
                            observer=MinMaxObserver(q_level="L", out_channels=None),
                            activation_weight_flag=0,
                            qaft=qaft,
                        )
                else:
                    if q_level == 0:
                        self.weight_quantizer = SymmetricQuantizer(
                            bits=w_bits,
                            observer=MovingAverageMinMaxObserver(
                                q_level="C", out_channels=out_channels
                            ),
                            activation_weight_flag=0,
                            qaft=qaft,
                        )
                    else:
                        self.weight_quantizer = SymmetricQuantizer(
                            bits=w_bits,
                            observer=MovingAverageMinMaxObserver(
                                q_level="L", out_channels=None
                            ),
                            activation_weight_flag=0,
                            qaft=qaft,
                        )
            else:
                self.activation_quantizer = AsymmetricQuantizer(
                    bits=a_bits,
                    observer=MovingAverageMinMaxObserver(
                        q_level="L", out_channels=None
                    ),
                    activation_weight_flag=1,
                    qaft=qaft,
                )
                if weight_observer == 0:
                    if q_level == 0:
                        self.weight_quantizer = AsymmetricQuantizer(
                            bits=w_bits,
                            observer=MinMaxObserver(
                                q_level="C", out_channels=out_channels
                            ),
                            activation_weight_flag=0,
                            qaft=qaft,
                        )
                    else:
                        self.weight_quantizer = AsymmetricQuantizer(
                            bits=w_bits,
                            observer=MinMaxObserver(q_level="L", out_channels=None),
                            activation_weight_flag=0,
                            qaft=qaft,
                        )
                else:
                    if q_level == 0:
                        self.weight_quantizer = AsymmetricQuantizer(
                            bits=w_bits,
                            observer=MovingAverageMinMaxObserver(
                                q_level="C", out_channels=out_channels
                            ),
                            activation_weight_flag=0,
                            qaft=qaft,
                        )
                    else:
                        self.weight_quantizer = AsymmetricQuantizer(
                            bits=w_bits,
                            observer=MovingAverageMinMaxObserver(
                                q_level="L", out_channels=None
                            ),
                            activation_weight_flag=0,
                            qaft=qaft,
                        )
        else:
            self.activation_quantizer = SymmetricQuantizer(
                bits=a_bits,
                observer=HistogramObserver(q_level="L", percentile=percentile),
                activation_weight_flag=1,
                qaft=qaft,
            )
            if weight_observer == 0:
                if q_level == 0:
                    self.weight_quantizer = SymmetricQuantizer(
                        bits=w_bits,
                        observer=MinMaxObserver(q_level="C", out_channels=out_channels),
                        activation_weight_flag=0,
                        qaft=qaft,
                    )
                else:
                    self.weight_quantizer = SymmetricQuantizer(
                        bits=w_bits,
                        observer=MinMaxObserver(q_level="L", out_channels=None),
                        activation_weight_flag=0,
                        qaft=qaft,
                    )
            else:
                if q_level == 0:
                    self.weight_quantizer = SymmetricQuantizer(
                        bits=w_bits,
                        observer=MovingAverageMinMaxObserver(
                            q_level="C", out_channels=out_channels
                        ),
                        activation_weight_flag=0,
                        qaft=qaft,
                    )
                else:
                    self.weight_quantizer = SymmetricQuantizer(
                        bits=w_bits,
                        observer=MovingAverageMinMaxObserver(
                            q_level="L", out_channels=None
                        ),
                        activation_weight_flag=0,
                        qaft=qaft,
                    )

    def forward(self, input):
        if not self.qaft:
            # qat, calibrate bn_statis_para
            # 训练态
            if self.training:
                # 先做普通卷积得到A，以取得BN参数
                output = F.conv2d(
                    input,
                    self.weight,
                    self.bias,
                    self.stride,
                    self.padding,
                    self.dilation,
                    self.groups,
                )
                # 更新BN统计参数（batch和running）
                dims = [dim for dim in range(4) if dim != 1]
                batch_mean = torch.mean(output, dim=dims)
                batch_var = torch.var(output, dim=dims)
                with torch.no_grad():
                    if not self.pretrained_model:
                        if self.num_flag == 0:
                            self.num_flag += 1
                            running_mean = batch_mean
                            running_var = batch_var
                        else:
                            running_mean = (
                                1 - self.momentum
                            ) * self.running_mean + self.momentum * batch_mean
                            running_var = (
                                1 - self.momentum
                            ) * self.running_var + self.momentum * batch_var
                        self.running_mean.copy_(running_mean)
                        self.running_var.copy_(running_var)
                    else:
                        running_mean = (
                            1 - self.momentum
                        ) * self.running_mean + self.momentum * batch_mean
                        running_var = (
                            1 - self.momentum
                        ) * self.running_var + self.momentum * batch_var
                        self.running_mean.copy_(running_mean)
                        self.running_var.copy_(running_var)
                # bn融合
                if self.bias is not None:
                    bias_fused = reshape_to_bias(
                        self.beta
                        + (self.bias - batch_mean)
                        * (self.gamma / torch.sqrt(batch_var + self.eps))
                    )
                else:
                    bias_fused = reshape_to_bias(
                        self.beta
                        - batch_mean * (self.gamma / torch.sqrt(batch_var + self.eps))
                    )  # b融batch
                # bn融合不校准
                if not self.bn_fuse_calib:
                    weight_fused = self.weight * reshape_to_weight(
                        self.gamma / torch.sqrt(batch_var + self.eps)
                    )  # w融batch
                # bn融合校准
                else:
                    weight_fused = self.weight * reshape_to_weight(
                        self.gamma / torch.sqrt(self.running_var + self.eps)
                    )  # w融running
            # 测试态
            else:
                if self.bias is not None:
                    bias_fused = reshape_to_bias(
                        self.beta
                        + (self.bias - self.running_mean)
                        * (self.gamma / torch.sqrt(self.running_var + self.eps))
                    )
                else:
                    bias_fused = reshape_to_bias(
                        self.beta
                        - self.running_mean
                        * (self.gamma / torch.sqrt(self.running_var + self.eps))
                    )  # b融running
                weight_fused = self.weight * reshape_to_weight(
                    self.gamma / torch.sqrt(self.running_var + self.eps)
                )  # w融running
        else:
            # qaft, freeze bn_statis_para
            if self.bias is not None:
                bias_fused = reshape_to_bias(
                    self.beta
                    + (self.bias - self.running_mean)
                    * (self.gamma / torch.sqrt(self.running_var + self.eps))
                )
            else:
                bias_fused = reshape_to_bias(
                    self.beta
                    - self.running_mean
                    * (self.gamma / torch.sqrt(self.running_var + self.eps))
                )  # b融running
            weight_fused = self.weight * reshape_to_weight(
                self.gamma / torch.sqrt(self.running_var + self.eps)
            )  # w融running

        # 量化A和bn融合后的W
        quant_input = self.activation_quantizer(input)
        quant_weight = self.weight_quantizer(weight_fused)

        if not self.qaft:
            # qat, quant_bn_fuse_conv
            # 量化卷积
            if self.training:  # 训练态
                # bn融合不校准
                if not self.bn_fuse_calib:
                    output = F.conv2d(
                        quant_input,
                        quant_weight,
                        bias_fused,
                        self.stride,
                        self.padding,
                        self.dilation,
                        self.groups,
                    )
                # bn融合校准
                else:
                    output = F.conv2d(
                        quant_input,
                        quant_weight,
                        None,
                        self.stride,
                        self.padding,
                        self.dilation,
                        self.groups,
                    )  # 注意，这里不加bias（self.bias为None）
                    # （这里将训练态下，卷积中w融合running参数的效果转为融合batch参数的效果）running ——> batch
                    output *= reshape_to_activation(
                        torch.sqrt(self.running_var + self.eps)
                        / torch.sqrt(batch_var + self.eps)
                    )
                    output += reshape_to_activation(bias_fused)
            else:  # 测试态
                output = F.conv2d(
                    quant_input,
                    quant_weight,
                    bias_fused,
                    self.stride,
                    self.padding,
                    self.dilation,
                    self.groups,
                )  # 注意，这里加bias，做完整的conv+bn
        else:
            # qaft, quant_bn_fuse_conv
            output = F.conv2d(
                quant_input,
                quant_weight,
                bias_fused,
                self.stride,
                self.padding,
                self.dilation,
                self.groups,
            )
        return output


class QuantLinear(nn.Linear):
    def __init__(
        self,
        in_features,
        out_features,
        bias=True,
        a_bits=8,
        w_bits=8,
        q_type=0,
        q_level=0,
        weight_observer=0,
        quant_inference=False,
        qaft=False,
        ptq=False,
        percentile=0.9999,
    ):
        super(QuantLinear, self).__init__(in_features, out_features, bias)
        self.quant_inference = quant_inference
        if not ptq:
            if q_type == 0:
                self.activation_quantizer = SymmetricQuantizer(
                    bits=a_bits,
                    observer=MovingAverageMinMaxObserver(
                        q_level="L", out_channels=None
                    ),
                    activation_weight_flag=1,
                    qaft=qaft,
                )
                if weight_observer == 0:
                    if q_level == 0:
                        self.weight_quantizer = SymmetricQuantizer(
                            bits=w_bits,
                            observer=MinMaxObserver(
                                q_level="FC", out_channels=out_features
                            ),
                            activation_weight_flag=0,
                            qaft=qaft,
                        )
                    else:
                        self.weight_quantizer = SymmetricQuantizer(
                            bits=w_bits,
                            observer=MinMaxObserver(q_level="L", out_channels=None),
                            activation_weight_flag=0,
                            qaft=qaft,
                        )
                else:
                    if q_level == 0:
                        self.weight_quantizer = SymmetricQuantizer(
                            bits=w_bits,
                            observer=MovingAverageMinMaxObserver(
                                q_level="FC", out_channels=out_features
                            ),
                            activation_weight_flag=0,
                            qaft=qaft,
                        )
                    else:
                        self.weight_quantizer = SymmetricQuantizer(
                            bits=w_bits,
                            observer=MovingAverageMinMaxObserver(
                                q_level="L", out_channels=None
                            ),
                            activation_weight_flag=0,
                            qaft=qaft,
                        )
            else:
                self.activation_quantizer = AsymmetricQuantizer(
                    bits=a_bits,
                    observer=MovingAverageMinMaxObserver(
                        q_level="L", out_channels=None
                    ),
                    activation_weight_flag=1,
                    qaft=qaft,
                )
                if weight_observer == 0:
                    if q_level == 0:
                        self.weight_quantizer = AsymmetricQuantizer(
                            bits=w_bits,
                            observer=MinMaxObserver(
                                q_level="FC", out_channels=out_features
                            ),
                            activation_weight_flag=0,
                            qaft=qaft,
                        )
                    else:
                        self.weight_quantizer = AsymmetricQuantizer(
                            bits=w_bits,
                            observer=MinMaxObserver(q_level="L", out_channels=None),
                            activation_weight_flag=0,
                            qaft=qaft,
                        )
                else:
                    if q_level == 0:
                        self.weight_quantizer = AsymmetricQuantizer(
                            bits=w_bits,
                            observer=MovingAverageMinMaxObserver(
                                q_level="FC", out_channels=out_features
                            ),
                            activation_weight_flag=0,
                            qaft=qaft,
                        )
                    else:
                        self.weight_quantizer = AsymmetricQuantizer(
                            bits=w_bits,
                            observer=MovingAverageMinMaxObserver(
                                q_level="L", out_channels=None
                            ),
                            activation_weight_flag=0,
                            qaft=qaft,
                        )
        else:
            self.activation_quantizer = SymmetricQuantizer(
                bits=a_bits,
                observer=HistogramObserver(q_level="L", percentile=percentile),
                activation_weight_flag=1,
                qaft=qaft,
            )
            if weight_observer == 0:
                if q_level == 0:
                    self.weight_quantizer = SymmetricQuantizer(
                        bits=w_bits,
                        observer=MinMaxObserver(
                            q_level="FC", out_channels=out_features
                        ),
                        activation_weight_flag=0,
                        qaft=qaft,
                    )
                else:
                    self.weight_quantizer = SymmetricQuantizer(
                        bits=w_bits,
                        observer=MinMaxObserver(q_level="L", out_channels=None),
                        activation_weight_flag=0,
                        qaft=qaft,
                    )
            else:
                if q_level == 0:
                    self.weight_quantizer = SymmetricQuantizer(
                        bits=w_bits,
                        observer=MovingAverageMinMaxObserver(
                            q_level="FC", out_channels=out_features
                        ),
                        activation_weight_flag=0,
                        qaft=qaft,
                    )
                else:
                    self.weight_quantizer = SymmetricQuantizer(
                        bits=w_bits,
                        observer=MovingAverageMinMaxObserver(
                            q_level="L", out_channels=None
                        ),
                        activation_weight_flag=0,
                        qaft=qaft,
                    )

    def forward(self, input):
        quant_input = self.activation_quantizer(input)
        if not self.quant_inference:
            quant_weight = self.weight_quantizer(self.weight)
        else:
            quant_weight = self.weight
        output = F.linear(quant_input, quant_weight, self.bias)
        return output


class QuantReLU(nn.ReLU):
    def __init__(
        self,
        inplace=False,
        a_bits=8,
        q_type=0,
        qaft=False,
        ptq=False,
        percentile=0.9999,
    ):
        super(QuantReLU, self).__init__(inplace)
        if not ptq:
            if q_type == 0:
                self.activation_quantizer = SymmetricQuantizer(
                    bits=a_bits,
                    observer=MovingAverageMinMaxObserver(
                        q_level="L", out_channels=None
                    ),
                    activation_weight_flag=1,
                    qaft=qaft,
                )
            else:
                self.activation_quantizer = AsymmetricQuantizer(
                    bits=a_bits,
                    observer=MovingAverageMinMaxObserver(
                        q_level="L", out_channels=None
                    ),
                    activation_weight_flag=1,
                    qaft=qaft,
                )
        else:
            self.activation_quantizer = SymmetricQuantizer(
                bits=a_bits,
                observer=HistogramObserver(q_level="L", percentile=percentile),
                activation_weight_flag=1,
                qaft=qaft,
            )

    def forward(self, input):
        quant_input = self.activation_quantizer(input)
        output = F.relu(quant_input, self.inplace)
        return output


class QuantLeakyReLU(nn.LeakyReLU):
    def __init__(
        self,
        negative_slope=0.01,
        inplace=False,
        a_bits=8,
        q_type=0,
        qaft=False,
        ptq=False,
        percentile=0.9999,
    ):
        super(QuantLeakyReLU, self).__init__(negative_slope, inplace)
        if not ptq:
            if q_type == 0:
                self.activation_quantizer = SymmetricQuantizer(
                    bits=a_bits,
                    observer=MovingAverageMinMaxObserver(
                        q_level="L", out_channels=None
                    ),
                    activation_weight_flag=1,
                    qaft=qaft,
                )
            else:
                self.activation_quantizer = AsymmetricQuantizer(
                    bits=a_bits,
                    observer=MovingAverageMinMaxObserver(
                        q_level="L", out_channels=None
                    ),
                    activation_weight_flag=1,
                    qaft=qaft,
                )
        else:
            self.activation_quantizer = SymmetricQuantizer(
                bits=a_bits,
                observer=HistogramObserver(q_level="L", percentile=percentile),
                activation_weight_flag=1,
                qaft=qaft,
            )

    def forward(self, input):
        quant_input = self.activation_quantizer(input)
        output = F.leaky_relu(quant_input, self.negative_slope, self.inplace)
        return output


class QuantSigmoid(nn.Sigmoid):
    def __init__(self, a_bits=8, q_type=0, qaft=False, ptq=False, percentile=0.9999):
        super(QuantSigmoid, self).__init__()
        if not ptq:
            if q_type == 0:
                self.activation_quantizer = SymmetricQuantizer(
                    bits=a_bits,
                    observer=MovingAverageMinMaxObserver(
                        q_level="L", out_channels=None
                    ),
                    activation_weight_flag=1,
                    qaft=qaft,
                )
            else:
                self.activation_quantizer = AsymmetricQuantizer(
                    bits=a_bits,
                    observer=MovingAverageMinMaxObserver(
                        q_level="L", out_channels=None
                    ),
                    activation_weight_flag=1,
                    qaft=qaft,
                )
        else:
            self.activation_quantizer = SymmetricQuantizer(
                bits=a_bits,
                observer=HistogramObserver(q_level="L", percentile=percentile),
                activation_weight_flag=1,
                qaft=qaft,
            )

    def forward(self, input):
        quant_input = self.activation_quantizer(input)
        output = F.sigmoid(quant_input)
        return output


class QuantMaxPool2d(nn.MaxPool2d):
    def __init__(
        self,
        kernel_size,
        stride=None,
        padding=0,
        dilation=1,
        return_indices=False,
        ceil_mode=False,
        a_bits=8,
        q_type=0,
        qaft=False,
        ptq=False,
        percentile=0.9999,
    ):
        super(QuantMaxPool2d, self).__init__(
            kernel_size, stride, padding, dilation, return_indices, ceil_mode
        )
        if not ptq:
            if q_type == 0:
                self.activation_quantizer = SymmetricQuantizer(
                    bits=a_bits,
                    observer=MovingAverageMinMaxObserver(
                        q_level="L", out_channels=None
                    ),
                    activation_weight_flag=1,
                    qaft=qaft,
                )
            else:
                self.activation_quantizer = AsymmetricQuantizer(
                    bits=a_bits,
                    observer=MovingAverageMinMaxObserver(
                        q_level="L", out_channels=None
                    ),
                    activation_weight_flag=1,
                    qaft=qaft,
                )
        else:
            self.activation_quantizer = SymmetricQuantizer(
                bits=a_bits,
                observer=HistogramObserver(q_level="L", percentile=percentile),
                activation_weight_flag=1,
                qaft=qaft,
            )

    def forward(self, input):
        quant_input = self.activation_quantizer(input)
        output = F.max_pool2d(
            quant_input,
            self.kernel_size,
            self.stride,
            self.padding,
            self.dilation,
            self.return_indices,
            self.ceil_mode,
        )
        return output


class QuantAvgPool2d(nn.AvgPool2d):
    def __init__(
        self,
        kernel_size,
        stride=None,
        padding=0,
        ceil_mode=False,
        count_include_pad=True,
        divisor_override=None,
        a_bits=8,
        q_type=0,
        qaft=False,
        ptq=False,
        percentile=0.9999,
    ):
        super(QuantAvgPool2d, self).__init__(
            kernel_size, stride, padding, ceil_mode, count_include_pad, divisor_override
        )
        if not ptq:
            if q_type == 0:
                self.activation_quantizer = SymmetricQuantizer(
                    bits=a_bits,
                    observer=MovingAverageMinMaxObserver(
                        q_level="L", out_channels=None
                    ),
                    activation_weight_flag=1,
                    qaft=qaft,
                )
            else:
                self.activation_quantizer = AsymmetricQuantizer(
                    bits=a_bits,
                    observer=MovingAverageMinMaxObserver(
                        q_level="L", out_channels=None
                    ),
                    activation_weight_flag=1,
                    qaft=qaft,
                )
        else:
            self.activation_quantizer = SymmetricQuantizer(
                bits=a_bits,
                observer=HistogramObserver(q_level="L", percentile=percentile),
                activation_weight_flag=1,
                qaft=qaft,
            )

    def forward(self, input):
        quant_input = self.activation_quantizer(input)
        output = F.avg_pool2d(
            quant_input,
            self.kernel_size,
            self.stride,
            self.padding,
            self.ceil_mode,
            self.count_include_pad,
            self.divisor_override,
        )
        return output


class QuantAdaptiveAvgPool2d(nn.AdaptiveAvgPool2d):
    def __init__(
        self, output_size, a_bits=8, q_type=0, qaft=False, ptq=False, percentile=0.9999
    ):
        super(QuantAdaptiveAvgPool2d, self).__init__(output_size)
        if not ptq:
            if q_type == 0:
                self.activation_quantizer = SymmetricQuantizer(
                    bits=a_bits,
                    observer=MovingAverageMinMaxObserver(
                        q_level="L", out_channels=None
                    ),
                    activation_weight_flag=1,
                    qaft=qaft,
                )
            else:
                self.activation_quantizer = AsymmetricQuantizer(
                    bits=a_bits,
                    observer=MovingAverageMinMaxObserver(
                        q_level="L", out_channels=None
                    ),
                    activation_weight_flag=1,
                    qaft=qaft,
                )
        else:
            self.activation_quantizer = SymmetricQuantizer(
                bits=a_bits,
                observer=HistogramObserver(q_level="L", percentile=percentile),
                activation_weight_flag=1,
                qaft=qaft,
            )

    def forward(self, input):
        quant_input = self.activation_quantizer(input)
        output = F.adaptive_avg_pool2d(quant_input, self.output_size)
        return output


class QuantAdd(nn.Module):
    def __init__(self, a_bits=8, q_type=0, qaft=False, ptq=False, percentile=0.9999):
        super(QuantAdd, self).__init__()
        if not ptq:
            self.observer_res = MovingAverageMinMaxObserver(
                q_level="L", out_channels=None
            )
            self.observer_shortcut = MovingAverageMinMaxObserver(
                q_level="L", out_channels=None
            )
            if q_type == 0:
                self.activation_quantizer = SymmetricQuantizer(
                    bits=a_bits,
                    observer=MovingAverageMinMaxObserver(
                        q_level="L", out_channels=None
                    ),
                    activation_weight_flag=1,
                    qaft=qaft,
                    union=True,
                )
            else:
                self.activation_quantizer = AsymmetricQuantizer(
                    bits=a_bits,
                    observer=MovingAverageMinMaxObserver(
                        q_level="L", out_channels=None
                    ),
                    activation_weight_flag=1,
                    qaft=qaft,
                    union=True,
                )
        else:
            self.observer_res = HistogramObserver(q_level="L", percentile=percentile)
            self.observer_shortcut = HistogramObserver(
                q_level="L", percentile=percentile
            )
            self.activation_quantizer = SymmetricQuantizer(
                bits=a_bits,
                observer=HistogramObserver(q_level="L", percentile=percentile),
                activation_weight_flag=1,
                qaft=qaft,
                union=True,
            )

    def forward(self, res, shortcut):
        self.observer_res(res)
        self.observer_shortcut(shortcut)
        observer_min_val = torch.min(
            self.observer_res.min_val, self.observer_shortcut.min_val
        )
        observer_max_val = torch.max(
            self.observer_res.max_val, self.observer_shortcut.max_val
        )
        self.activation_quantizer.observer.min_val = observer_min_val
        self.activation_quantizer.observer.max_val = observer_max_val
        quant_res = self.activation_quantizer(res)
        quant_shortcut = self.activation_quantizer(shortcut)
        output = quant_res + quant_shortcut
        return output


def add_quant_op(
    module,
    a_bits=8,
    w_bits=8,
    q_type=0,
    q_level=0,
    weight_observer=0,
    bn_fuse=False,
    bn_fuse_calib=False,
    quant_inference=False,
    pretrained_model=False,
    qaft=False,
    ptq=False,
    percentile=0.9999,
):
    for name, child in module.named_children():
        if isinstance(child, nn.Conv2d):
            if bn_fuse:
                conv_name_temp = name
                conv_child_temp = child
            else:
                if child.bias is not None:
                    quant_conv = QuantConv2d(
                        child.in_channels,
                        child.out_channels,
                        child.kernel_size,
                        stride=child.stride,
                        padding=child.padding,
                        dilation=child.dilation,
                        groups=child.groups,
                        bias=True,
                        padding_mode=child.padding_mode,
                        a_bits=a_bits,
                        w_bits=w_bits,
                        q_type=q_type,
                        q_level=q_level,
                        weight_observer=weight_observer,
                        quant_inference=quant_inference,
                        qaft=qaft,
                        ptq=ptq,
                        percentile=percentile,
                    )
                    quant_conv.bias.data = child.bias
                else:
                    quant_conv = QuantConv2d(
                        child.in_channels,
                        child.out_channels,
                        child.kernel_size,
                        stride=child.stride,
                        padding=child.padding,
                        dilation=child.dilation,
                        groups=child.groups,
                        bias=False,
                        padding_mode=child.padding_mode,
                        a_bits=a_bits,
                        w_bits=w_bits,
                        q_type=q_type,
                        q_level=q_level,
                        weight_observer=weight_observer,
                        quant_inference=quant_inference,
                        qaft=qaft,
                        ptq=ptq,
                        percentile=percentile,
                    )
                quant_conv.weight.data = child.weight
                module._modules[name] = quant_conv
        elif isinstance(child, nn.BatchNorm2d):
            if bn_fuse:
                if conv_child_temp.bias is not None:
                    quant_bn_fuse_conv = QuantBNFuseConv2d(
                        conv_child_temp.in_channels,
                        conv_child_temp.out_channels,
                        conv_child_temp.kernel_size,
                        stride=conv_child_temp.stride,
                        padding=conv_child_temp.padding,
                        dilation=conv_child_temp.dilation,
                        groups=conv_child_temp.groups,
                        bias=True,
                        padding_mode=conv_child_temp.padding_mode,
                        eps=child.eps,
                        momentum=child.momentum,
                        a_bits=a_bits,
                        w_bits=w_bits,
                        q_type=q_type,
                        q_level=q_level,
                        weight_observer=weight_observer,
                        pretrained_model=pretrained_model,
                        qaft=qaft,
                        ptq=ptq,
                        percentile=percentile,
                        bn_fuse_calib=bn_fuse_calib,
                    )
                    quant_bn_fuse_conv.bias.data = conv_child_temp.bias
                else:
                    quant_bn_fuse_conv = QuantBNFuseConv2d(
                        conv_child_temp.in_channels,
                        conv_child_temp.out_channels,
                        conv_child_temp.kernel_size,
                        stride=conv_child_temp.stride,
                        padding=conv_child_temp.padding,
                        dilation=conv_child_temp.dilation,
                        groups=conv_child_temp.groups,
                        bias=False,
                        padding_mode=conv_child_temp.padding_mode,
                        eps=child.eps,
                        momentum=child.momentum,
                        a_bits=a_bits,
                        w_bits=w_bits,
                        q_type=q_type,
                        q_level=q_level,
                        weight_observer=weight_observer,
                        pretrained_model=pretrained_model,
                        qaft=qaft,
                        ptq=ptq,
                        percentile=percentile,
                        bn_fuse_calib=bn_fuse_calib,
                    )
                quant_bn_fuse_conv.weight.data = conv_child_temp.weight
                quant_bn_fuse_conv.gamma.data = child.weight
                quant_bn_fuse_conv.beta.data = child.bias
                quant_bn_fuse_conv.running_mean.copy_(child.running_mean)
                quant_bn_fuse_conv.running_var.copy_(child.running_var)
                module._modules[conv_name_temp] = quant_bn_fuse_conv
                module._modules[name] = nn.Identity()
        elif isinstance(child, nn.ConvTranspose2d):
            if child.bias is not None:
                quant_conv_transpose = QuantConvTranspose2d(
                    child.in_channels,
                    child.out_channels,
                    child.kernel_size,
                    stride=child.stride,
                    padding=child.padding,
                    output_padding=child.output_padding,
                    groups=child.groups,
                    bias=True,
                    dilation=child.dilation,
                    padding_mode=child.padding_mode,
                    a_bits=a_bits,
                    w_bits=w_bits,
                    q_type=q_type,
                    weight_observer=weight_observer,
                    quant_inference=quant_inference,
                    qaft=qaft,
                    ptq=ptq,
                    percentile=percentile,
                )
                quant_conv_transpose.bias.data = child.bias
            else:
                quant_conv_transpose = QuantConvTranspose2d(
                    child.in_channels,
                    child.out_channels,
                    child.kernel_size,
                    stride=child.stride,
                    padding=child.padding,
                    output_padding=child.output_padding,
                    groups=child.groups,
                    bias=False,
                    dilation=child.dilation,
                    padding_mode=child.padding_mode,
                    a_bits=a_bits,
                    w_bits=w_bits,
                    q_type=q_type,
                    weight_observer=weight_observer,
                    quant_inference=quant_inference,
                    qaft=qaft,
                    ptq=ptq,
                    percentile=percentile,
                )
            quant_conv_transpose.weight.data = child.weight
            module._modules[name] = quant_conv_transpose
        elif isinstance(child, nn.Linear):
            if child.bias is not None:
                quant_linear = QuantLinear(
                    child.in_features,
                    child.out_features,
                    bias=True,
                    a_bits=a_bits,
                    w_bits=w_bits,
                    q_type=q_type,
                    q_level=q_level,
                    weight_observer=weight_observer,
                    quant_inference=quant_inference,
                    qaft=qaft,
                    ptq=ptq,
                    percentile=percentile,
                )
                quant_linear.bias.data = child.bias
            else:
                quant_linear = QuantLinear(
                    child.in_features,
                    child.out_features,
                    bias=False,
                    a_bits=a_bits,
                    w_bits=w_bits,
                    q_type=q_type,
                    q_level=q_level,
                    weight_observer=weight_observer,
                    quant_inference=quant_inference,
                    qaft=qaft,
                    ptq=ptq,
                    percentile=percentile,
                )
            quant_linear.weight.data = child.weight
            module._modules[name] = quant_linear
        # relu needn’t quantize, it will be fused in quant_inference
        # elif isinstance(child, nn.ReLU):
        #    quant_relu = QuantReLU(inplace=child.inplace, a_bits=a_bits,
        #                           q_type=q_type, qaft=qaft, ptq=ptq, percentile=percentile)
        #    module._modules[name] = quant_relu
        elif isinstance(child, nn.LeakyReLU):
            quant_leaky_relu = QuantLeakyReLU(
                negative_slope=child.negative_slope,
                inplace=child.inplace,
                a_bits=a_bits,
                q_type=q_type,
                qaft=qaft,
                ptq=ptq,
                percentile=percentile,
            )
            module._modules[name] = quant_leaky_relu
        elif isinstance(child, nn.Sigmoid):
            quant_sigmoid = QuantSigmoid(
                a_bits=a_bits, q_type=q_type, qaft=qaft, ptq=ptq, percentile=percentile
            )
            module._modules[name] = quant_sigmoid
        elif isinstance(child, nn.MaxPool2d):
            quant_max_pool = QuantMaxPool2d(
                kernel_size=child.kernel_size,
                stride=child.stride,
                padding=child.padding,
                a_bits=a_bits,
                q_type=q_type,
                qaft=qaft,
                ptq=ptq,
                percentile=percentile,
            )
            module._modules[name] = quant_max_pool
        elif isinstance(child, nn.AvgPool2d):
            quant_avg_pool = QuantAvgPool2d(
                kernel_size=child.kernel_size,
                stride=child.stride,
                padding=child.padding,
                a_bits=a_bits,
                q_type=q_type,
                qaft=qaft,
                ptq=ptq,
                percentile=percentile,
            )
            module._modules[name] = quant_avg_pool
        elif isinstance(child, nn.AdaptiveAvgPool2d):
            quant_adaptive_avg_pool = QuantAdaptiveAvgPool2d(
                output_size=child.output_size,
                a_bits=a_bits,
                q_type=q_type,
                qaft=qaft,
                ptq=ptq,
                percentile=percentile,
            )
            module._modules[name] = quant_adaptive_avg_pool
        elif isinstance(child, Add):
            quant_add = QuantAdd(
                a_bits=a_bits, q_type=q_type, qaft=qaft, ptq=ptq, percentile=percentile
            )
            module._modules[name] = quant_add
        # elif isinstance(child, Concat):
        #    quant_concat = QuantConcat(dim=child.dim,
        #                               a_bits=a_bits,
        #                               q_type=q_type,
        #                               qaft=qaft,
        #                               ptq=ptq,
        #                               percentile=percentile)
        #    module._modules[name] = quant_concat
        else:
            add_quant_op(
                child,
                a_bits=a_bits,
                w_bits=w_bits,
                q_type=q_type,
                q_level=q_level,
                weight_observer=weight_observer,
                bn_fuse=bn_fuse,
                bn_fuse_calib=bn_fuse_calib,
                quant_inference=quant_inference,
                pretrained_model=pretrained_model,
                qaft=qaft,
                ptq=ptq,
                percentile=percentile,
            )


def prepare(
    model,
    inplace=False,
    a_bits=8,
    w_bits=8,
    q_type=0,
    q_level=0,
    weight_observer=0,
    bn_fuse=False,
    bn_fuse_calib=False,
    quant_inference=False,
    pretrained_model=False,
    qaft=False,
    ptq=False,
    percentile=0.9999,
):
    if not inplace:
        model = copy.deepcopy(model)
    add_quant_op(
        model,
        a_bits=a_bits,
        w_bits=w_bits,
        q_type=q_type,
        q_level=q_level,
        weight_observer=weight_observer,
        bn_fuse=bn_fuse,
        bn_fuse_calib=bn_fuse_calib,
        quant_inference=quant_inference,
        pretrained_model=pretrained_model,
        qaft=qaft,
        ptq=ptq,
        percentile=percentile,
    )
    return model


# *** temp_dev ***
"""
class QuantConcat(nn.Module):
    def __init__(self, dim=1, a_bits=8, q_type=0, qaft=False, ptq=False, percentile=0.9999):
        super(QuantConcat, self).__init__()
        self.dim = dim
        if not ptq:
            self.observer_res = MovingAverageMinMaxObserver(q_level='L', out_channels=None)
            self.observer_shortcut = MovingAverageMinMaxObserver(q_level='L', out_channels=None)
            if q_type == 0:
                self.activation_quantizer = SymmetricQuantizer(bits=a_bits, observer=MovingAverageMinMaxObserver(
                                                               q_level='L', out_channels=None), activation_weight_flag=1, qaft=qaft, union=True)
            else:
                self.activation_quantizer = AsymmetricQuantizer(bits=a_bits, observer=MovingAverageMinMaxObserver(
                                                                q_level='L', out_channels=None), activation_weight_flag=1, qaft=qaft, union=True)
        else:
            self.observer_res = HistogramObserver(q_level='L', percentile=percentile)
            self.observer_shortcut = HistogramObserver(q_level='L', percentile=percentile)
            self.activation_quantizer = SymmetricQuantizer(bits=a_bits, observer=HistogramObserver(
                                                           q_level='L', percentile=percentile), activation_weight_flag=1, qaft=qaft, union=True)

    def forward(self, res, shortcut):
        self.observer_res(res)
        self.observer_shortcut(shortcut)
        observer_min_val = torch.min(self.observer_res.min_val, self.observer_shortcut.min_val)
        observer_max_val = torch.max(self.observer_res.max_val, self.observer_shortcut.max_val)
        self.activation_quantizer.observer.min_val = observer_min_val
        self.activation_quantizer.observer.max_val = observer_max_val
        quant_res = self.activation_quantizer(res)
        quant_shortcut = self.activation_quantizer(shortcut)
        output = torch.cat([quant_shortcut, quant_res], dim=self.dim)
        return output
"""



import torch
import torch.nn as nn

from micronet.base_module.op import *


class BasicBlock(nn.Module):
    """Basic Block for resnet 18 and resnet 34"""

    # BasicBlock and BottleNeck block
    # have different output size
    # we use class attribute expansion
    # to distinct
    expansion = 1

    def __init__(self, in_channels, out_channels, stride=1):
        super().__init__()

        # residual function
        self.residual_function = nn.Sequential(
            nn.Conv2d(
                in_channels,
                out_channels,
                kernel_size=3,
                stride=stride,
                padding=1,
                bias=False,
            ),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(inplace=True),
            nn.Conv2d(
                out_channels,
                out_channels * BasicBlock.expansion,
                kernel_size=3,
                padding=1,
                bias=False,
            ),
            nn.BatchNorm2d(out_channels * BasicBlock.expansion),
        )

        # shortcut
        self.shortcut = nn.Sequential()

        # the shortcut output dimension is not the same with residual function
        # use 1*1 convolution to match the dimension
        if stride != 1 or in_channels != BasicBlock.expansion * out_channels:
            self.shortcut = nn.Sequential(
                nn.Conv2d(
                    in_channels,
                    out_channels * BasicBlock.expansion,
                    kernel_size=1,
                    stride=stride,
                    bias=False,
                ),
                nn.BatchNorm2d(out_channels * BasicBlock.expansion),
            )
        # +++
        self.add = Add()

    def forward(self, x):
        # return nn.ReLU(inplace=True)(self.residual_function(x) + self.shortcut(x))
        # +++
        return nn.ReLU(inplace=True)(
            self.add(self.residual_function(x), self.shortcut(x))
        )


class BottleNeck(nn.Module):
    """Residual block for resnet over 50 layers"""

    expansion = 4

    def __init__(self, in_channels, out_channels, stride=1):
        super().__init__()
        self.residual_function = nn.Sequential(
            nn.Conv2d(in_channels, out_channels, kernel_size=1, bias=False),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(inplace=True),
            nn.Conv2d(
                out_channels,
                out_channels,
                stride=stride,
                kernel_size=3,
                padding=1,
                bias=False,
            ),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(inplace=True),
            nn.Conv2d(
                out_channels,
                out_channels * BottleNeck.expansion,
                kernel_size=1,
                bias=False,
            ),
            nn.BatchNorm2d(out_channels * BottleNeck.expansion),
        )

        self.shortcut = nn.Sequential()

        if stride != 1 or in_channels != out_channels * BottleNeck.expansion:
            self.shortcut = nn.Sequential(
                nn.Conv2d(
                    in_channels,
                    out_channels * BottleNeck.expansion,
                    stride=stride,
                    kernel_size=1,
                    bias=False,
                ),
                nn.BatchNorm2d(out_channels * BottleNeck.expansion),
            )
        # +++
        self.add = Add()

    def forward(self, x):
        # return nn.ReLU(inplace=True)(self.residual_function(x) + self.shortcut(x))
        # +++
        return nn.ReLU(inplace=True)(
            self.add(self.residual_function(x), self.shortcut(x))
        )


class ResNet(nn.Module):
    def __init__(self, block, num_block, num_classes=10):
        super().__init__()

        self.in_channels = 64

        self.conv1 = nn.Sequential(
            nn.Conv2d(3, 64, kernel_size=3, padding=1, bias=False),
            nn.BatchNorm2d(64),
            nn.ReLU(inplace=True),
        )
        # we use a different inputsize than the original paper
        # so conv2_x's stride is 1
        self.conv2_x = self._make_layer(block, 64, num_block[0], 1)
        self.conv3_x = self._make_layer(block, 128, num_block[1], 2)
        self.conv4_x = self._make_layer(block, 256, num_block[2], 2)
        self.conv5_x = self._make_layer(block, 512, num_block[3], 2)
        self.avg_pool = nn.AdaptiveAvgPool2d((1, 1))
        self.fc = nn.Linear(512 * block.expansion, num_classes)

    def _make_layer(self, block, out_channels, num_blocks, stride):
        """make resnet layers(by layer i didnt mean this 'layer' was the
        same as a neuron netowork layer, ex. conv layer), one layer may
        contain more than one residual block

        Args:
            block: block type, basic block or bottle neck block
            out_channels: output depth channel number of this layer
            num_blocks: how many blocks per layer
            stride: the stride of the first block of this layer

        Return:
            return a resnet layer
        """

        # we have num_block blocks per layer, the first block
        # could be 1 or 2, other blocks would always be 1
        strides = [stride] + [1] * (num_blocks - 1)
        layers = []
        for stride in strides:
            layers.append(block(self.in_channels, out_channels, stride))
            self.in_channels = out_channels * block.expansion

        return nn.Sequential(*layers)

    def forward(self, x):
        output = self.conv1(x)
        output = self.conv2_x(output)
        output = self.conv3_x(output)
        output = self.conv4_x(output)
        output = self.conv5_x(output)
        output = self.avg_pool(output)
        output = output.view(output.size(0), -1)
        output = self.fc(output)

        return output


def resnet18():
    """return a ResNet 18 object"""
    return ResNet(BasicBlock, [2, 2, 2, 2])


def resnet34():
    """return a ResNet 34 object"""
    return ResNet(BasicBlock, [3, 4, 6, 3])


def resnet50():
    """return a ResNet 50 object"""
    return ResNet(BottleNeck, [3, 4, 6, 3])


def resnet101():
    """return a ResNet 101 object"""
    return ResNet(BottleNeck, [3, 4, 23, 3])


def resnet152():
    """return a ResNet 152 object"""
    return ResNet(BottleNeck, [3, 8, 36, 3])



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import sys

sys.path.append("../..")
import math
import os
import argparse
import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim
import torchvision
import torchvision.transforms as transforms
from torch.autograd import Variable
from torch.nn import init
from models import nin_gc, nin


def setup_seed(seed):
    torch.manual_seed(seed)
    # torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    np.random.seed(seed)
    torch.backends.cudnn.deterministic = True


def save_state(model, best_acc):
    print("==> Saving model ...")
    state = {
        "best_acc": best_acc,
        "state_dict": model.state_dict(),
    }
    state_copy = state["state_dict"].copy()
    for key in state_copy.keys():
        if "module" in key:
            state["state_dict"][key.replace("module.", "")] = state["state_dict"].pop(
                key
            )
    if args.model_type == 0:
        if args.sr:
            torch.save(state, "models_save/nin_sparse.pth")
        elif args.prune_refine:
            torch.save(
                {"cfg": cfg, "best_acc": best_acc, "state_dict": state["state_dict"]},
                "models_save/nin_finetune.pth",
            )
        else:
            torch.save(state, "models_save/nin.pth")
    else:
        if args.sr:
            torch.save(state, "models_save/nin_gc_sparse.pth")
        elif args.gc_prune_refine:
            torch.save(
                {"cfg": cfg, "best_acc": best_acc, "state_dict": state["state_dict"]},
                "models_save/nin_gc_retrain.pth",
            )
        else:
            torch.save(state, "models_save/nin_gc.pth")


# ***********************稀疏训练（对BN层γ进行约束）**************************
def updateBN():
    for m in model.modules():
        if isinstance(m, nn.BatchNorm2d):
            if hasattr(m.weight, "data"):
                m.weight.grad.data.add_(args.s * torch.sign(m.weight.data))  # L1正则


def train(epoch):
    model.train()

    for batch_idx, (data, target) in enumerate(trainloader):
        if not args.cpu:
            data, target = data.cuda(), target.cuda()
        data, target = Variable(data), Variable(target)
        output = model(data)
        loss = criterion(output, target)

        optimizer.zero_grad()
        loss.backward()

        # ***********************稀疏训练（对BN层γ进行约束）**************************
        if args.sr:
            updateBN()

        optimizer.step()

        if batch_idx % 100 == 0:
            print(
                "Train Epoch: {} [{}/{} ({:.0f}%)]\tLoss: {:.6f}\tLR: {}".format(
                    epoch,
                    batch_idx * len(data),
                    len(trainloader.dataset),
                    100.0 * batch_idx / len(trainloader),
                    loss.data.item(),
                    optimizer.param_groups[0]["lr"],
                )
            )
    return


def test():
    global best_acc
    model.eval()
    test_loss = 0
    correct = 0

    for data, target in testloader:
        if not args.cpu:
            data, target = data.cuda(), target.cuda()
        data, target = Variable(data), Variable(target)
        output = model(data)
        test_loss += criterion(output, target).data.item()
        pred = output.data.max(1, keepdim=True)[1]
        correct += pred.eq(target.data.view_as(pred)).cpu().sum()
    acc = 100.0 * float(correct) / len(testloader.dataset)

    if acc > best_acc:
        best_acc = acc
        save_state(model, best_acc)
    average_test_loss = test_loss / (len(testloader.dataset) / args.eval_batch_size)

    print(
        "\nTest set: Average loss: {:.4f}, Accuracy: {}/{} ({:.2f}%)".format(
            average_test_loss,
            correct,
            len(testloader.dataset),
            100.0 * float(correct) / len(testloader.dataset),
        )
    )

    print("Best Accuracy: {:.2f}%\n".format(best_acc))
    return


def adjust_learning_rate(optimizer, epoch):
    update_list = [80, 130, 180, 230, 280]
    if epoch in update_list:
        for param_group in optimizer.param_groups:
            param_group["lr"] = param_group["lr"] * 0.1
    return


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--cpu", action="store_true", help="set if only CPU is available"
    )
    parser.add_argument("--gpu_id", action="store", default="", help="gpu_id")
    parser.add_argument(
        "--data", action="store", default="../../data", help="dataset path"
    )
    parser.add_argument(
        "--lr", action="store", default=0.01, help="the intial learning rate"
    )
    parser.add_argument("--wd", action="store", default=1e-7, help="nin_gc:0, nin:1e-5")
    # prune_refine
    parser.add_argument(
        "--prune_refine",
        default="",
        type=str,
        metavar="PATH",
        help="the path to the prune_refine model",
    )
    # refine
    parser.add_argument(
        "--refine",
        default="",
        type=str,
        metavar="PATH",
        help="the path to the float_refine model",
    )
    # resume
    parser.add_argument(
        "--resume",
        default="",
        type=str,
        metavar="PATH",
        help="the path to the resume model",
    )
    # gc_prune_refine的cfg
    parser.add_argument(
        "--gc_prune_refine", nargs="+", type=int, help="gc_prune_refine-cfg"
    )
    parser.add_argument("--train_batch_size", type=int, default=128)
    parser.add_argument("--eval_batch_size", type=int, default=256)
    parser.add_argument("--num_workers", type=int, default=2)
    parser.add_argument(
        "--epochs", type=int, default=300, metavar="N", help="number of epochs to train"
    )
    # sr(稀疏标志)
    parser.add_argument(
        "--sparsity-regularization",
        "-sr",
        dest="sr",
        action="store_true",
        help="train with channel sparsity regularization",
    )
    # s(稀疏率)
    parser.add_argument(
        "--s", type=float, default=0.0001, help="nin:0.0001, nin_gc:0.001"
    )
    parser.add_argument(
        "--model_type", type=int, default=1, help="model type:0-nin,1-nin_gc"
    )
    args = parser.parse_args()
    print("==> Options:", args)

    if args.gpu_id:
        os.environ["CUDA_VISIBLE_DEVICES"] = args.gpu_id

    setup_seed(1)

    print("==> Preparing data..")
    transform_train = transforms.Compose(
        [
            transforms.RandomCrop(32, padding=4),
            transforms.RandomHorizontalFlip(),
            transforms.ToTensor(),
            transforms.Normalize((0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010)),
        ]
    )

    transform_test = transforms.Compose(
        [
            transforms.ToTensor(),
            transforms.Normalize((0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010)),
        ]
    )

    trainset = torchvision.datasets.CIFAR10(
        root=args.data, train=True, download=True, transform=transform_train
    )
    trainloader = torch.utils.data.DataLoader(
        trainset,
        batch_size=args.train_batch_size,
        shuffle=True,
        num_workers=args.num_workers,
    )

    testset = torchvision.datasets.CIFAR10(
        root=args.data, train=False, download=True, transform=transform_test
    )
    testloader = torch.utils.data.DataLoader(
        testset,
        batch_size=args.eval_batch_size,
        shuffle=False,
        num_workers=args.num_workers,
    )

    classes = (
        "plane",
        "car",
        "bird",
        "cat",
        "deer",
        "dog",
        "frog",
        "horse",
        "ship",
        "truck",
    )

    if args.prune_refine:
        print("******Prune Refine model******")
        # checkpoint = torch.load('models_save/nin_prune.pth')
        checkpoint = torch.load(args.prune_refine)
        cfg = checkpoint["cfg"]
        model = nin.Net(cfg=checkpoint["cfg"])
        model.load_state_dict(checkpoint["state_dict"])
        best_acc = 0
    elif args.refine:
        print("******Float Refine model******")
        # checkpoint = torch.load('models_save/nin.pth')
        state_dict = torch.load(args.refine)
        if args.model_type == 0:
            model = nin.Net()
        else:
            model = nin_gc.Net()
        model.load_state_dict(state_dict)
        best_acc = 0
    elif args.resume:
        print("******Reume model******")
        # checkpoint = torch.load('models_save/nin.pth')
        # checkpoint = torch.load('models_save/nin_sparse.pth')
        checkpoint = torch.load(args.resume)
        if args.model_type == 0:
            model = nin.Net()
        else:
            model = nin_gc.Net()
        model.load_state_dict(checkpoint["state_dict"])
        best_acc = checkpoint["best_acc"]
    else:
        # nin_gc_retrain
        if args.gc_prune_refine:
            print("******GCPrune Refine model******")
            cfg = args.gc_prune_refine
            model = nin_gc.Net(cfg=cfg)
        else:
            print("******Initializing model******")
            if args.model_type == 0:
                model = nin.Net()
            else:
                model = nin_gc.Net()
        best_acc = 0
        for m in model.modules():
            if isinstance(m, nn.Conv2d):
                init.xavier_uniform_(m.weight)
                if m.bias is not None:
                    init.zeros_(m.bias)
            elif isinstance(m, nn.Linear):
                init.normal_(m.weight, 0, 0.01)
                if m.bias is not None:
                    init.zeros_(m.bias)

    if not args.cpu:
        model.cuda()
        model = torch.nn.DataParallel(
            model, device_ids=range(torch.cuda.device_count())
        )
    print(model)

    base_lr = float(args.lr)
    param_dict = dict(model.named_parameters())
    params = []

    for key, value in param_dict.items():
        params += [{"params": [value], "lr": base_lr, "weight_decay": args.wd}]

    criterion = nn.CrossEntropyLoss()
    optimizer = optim.Adam(params, lr=base_lr, weight_decay=args.wd)

    for epoch in range(1, args.epochs):
        adjust_learning_rate(optimizer, epoch)
        train(epoch)
        test()



import sys

sys.path.append("..")

import numpy as np
import torch
import torch.nn as nn
import torchvision
import math
import time

# tensorrt-lib
import util_trt


# tensorrt eval/test
class SegmentationModule_v2_trt(nn.Module):
    def __init__(
        self,
        context,
        buffers,
        crit,
        deep_sup_scale=None,
        use_softmax=False,
        binding_id=0,
    ):
        super(SegmentationModule_v2_trt, self).__init__()
        self.context = context
        self.inputs, self.outputs, self.bindings, self.stream = buffers
        self.crit = crit
        self.deep_sup_scale = deep_sup_scale
        self.use_softmax = use_softmax
        self.binding_id = binding_id

    def pixel_acc(self, pred, label):
        _, preds = torch.max(pred, dim=1)
        valid = (label >= 0).long()
        acc_sum = torch.sum(valid * (preds == label).long())
        pixel_sum = torch.sum(valid)
        acc = acc_sum.float() / (pixel_sum.float() + 1e-10)
        return acc

    # eval/test
    def forward(self, feed_dict, *, segSize=None, shape_of_input):
        shape_of_output = (1, 2, int(shape_of_input[2] / 8), int(shape_of_input[3] / 8))
        self.inputs[0].host = (
            util_trt.to_numpy(feed_dict["img_data"])
            .astype(dtype=np.float32)
            .reshape(-1)
        )
        trt_outputs, infer_time = util_trt.do_inference_v2(
            context=self.context,
            bindings=self.bindings,
            inputs=self.inputs,
            outputs=self.outputs,
            stream=self.stream,
            h_=shape_of_input[2],
            w_=shape_of_input[3],
            binding_id=self.binding_id,
        )
        trt_outputs = trt_outputs[0][
            : shape_of_output[0]
            * shape_of_output[1]
            * shape_of_output[2]
            * shape_of_output[3]
        ]
        results = util_trt.postprocess_the_outputs(trt_outputs, shape_of_output)
        x = torch.from_numpy(results)
        x = x.cuda()
        if self.use_softmax:
            x = nn.functional.interpolate(
                x, size=segSize, mode="bilinear", align_corners=False
            )
            x = nn.functional.softmax(x, dim=1)
        else:
            x = nn.functional.log_softmax(x, dim=1)
        pred = x
        return pred, infer_time


class C1_unet_v3(nn.Module):
    def __init__(self, num_class=150, fc_dim=2048, use_softmax=False):
        super(C1_unet_v3, self).__init__()
        self.use_softmax = use_softmax

        # self.cbr1 = conv3x3_bn_relu(fc_dim//8, fc_dim // 4, 2)
        self.cbr1 = nn.Sequential(
            conv3x3_bn_relu(fc_dim, fc_dim, 1),
            conv3x3_bn_relu(fc_dim, fc_dim, 1),
            conv3x3_bn_relu(fc_dim, fc_dim, 1),
        )

        self.cbr2 = nn.Sequential(
            conv3x3_bn_relu(fc_dim // 2, fc_dim // 2, 1),
            conv3x3_bn_relu(fc_dim // 2, fc_dim // 2, 1),
            conv3x3_bn_relu(fc_dim // 2, fc_dim // 2, 1),
        )

        self.cbr3 = conv3x3_bn_relu(fc_dim // 2 * 3, fc_dim // 2 * 3, 1)

        self.cbr4 = nn.Sequential(
            conv3x3_bn_relu(fc_dim // 4, fc_dim // 4, 1),
            conv3x3_bn_relu(fc_dim // 4, fc_dim // 4, 1),
            conv3x3_bn_relu(fc_dim // 4, fc_dim // 4, 1),
        )

        self.cbr5 = nn.Sequential(
            conv3x3_bn_relu(fc_dim // 4 * 7, fc_dim // 2, 1),
            conv3x3_bn_relu(fc_dim // 2, fc_dim // 2, 1),
        )

        # self.conv_second = nn.Conv2d(fc_dim // 4, fc_dim // 4, 3, 1, 0)
        self.conv_last = nn.Conv2d(fc_dim // 2, num_class, 1, 1, 0)

        # self.upsample = Upsample(scale=2, mode='bilinear', align_corners=False)

    def forward(self, conv_out, segSize=None):
        shape = list(conv_out[1].size())
        shape = shape[2:]
        shape2 = list(conv_out[2].size())
        shape2 = shape2[2:]
        # x1 = self.cbr1(conv_out[0])
        x2 = self.cbr1(conv_out[3])
        # x2 = nn.functional.interpolate(
        #    x2, size=shape2, mode='bilinear', align_corners=False)
        x2 = nn.functional.interpolate(
            x2,
            size=(int(shape2[0]), int(shape2[1])),
            mode="bilinear",
            align_corners=False,
        )  # +++
        x3 = self.cbr2(conv_out[2])
        x3 = torch.cat([x2, x3], 1)

        # x3 = nn.functional.interpolate(
        #    x3, size=shape, mode='bilinear', align_corners=False)
        x3 = nn.functional.interpolate(
            x3,
            size=(int(shape[0]), int(shape[1])),
            mode="bilinear",
            align_corners=False,
        )  # +++
        x3 = self.cbr3(x3)
        x4 = self.cbr4(conv_out[1])
        x4 = torch.cat([x4, x3], 1)

        x5 = self.cbr5(x4)
        x = self.conv_last(x5)
        """
        if self.use_softmax: # is True during inference
            x = nn.functional.interpolate(
                x, size=segSize, mode='bilinear', align_corners=False)
            x = nn.functional.softmax(x, dim=1)
        else:
            x = nn.functional.log_softmax(x, dim=1)
        """
        return x



import torch.nn as nn


class ConvBNReLU(nn.Module):
    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        stride=1,
        padding=0,
        dilation=1,
        groups=1,
        bias=True,
        padding_mode="zeros",
        eps=1e-5,
        momentum=0.1,
    ):
        super(ConvBNReLU, self).__init__()

        self.conv = nn.Conv2d(
            in_channels,
            out_channels,
            kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            groups=groups,
            bias=bias,
            padding_mode=padding_mode,
        )
        self.bn = nn.BatchNorm2d(out_channels, eps=eps, momentum=momentum)
        self.relu = nn.ReLU(inplace=True)

    def forward(self, x):
        x = self.conv(x)
        x = self.bn(x)
        x = self.relu(x)
        return x


class Net(nn.Module):
    def __init__(self, cfg=None):
        super(Net, self).__init__()
        if cfg is None:
            cfg = [192, 160, 96, 192, 192, 192, 192, 192]
        self.model = nn.Sequential(
            ConvBNReLU(3, cfg[0], kernel_size=5, stride=1, padding=2),
            ConvBNReLU(cfg[0], cfg[1], kernel_size=1, stride=1, padding=0),
            ConvBNReLU(cfg[1], cfg[2], kernel_size=1, stride=1, padding=0),
            nn.MaxPool2d(kernel_size=3, stride=2, padding=1),
            ConvBNReLU(cfg[2], cfg[3], kernel_size=5, stride=1, padding=2),
            ConvBNReLU(cfg[3], cfg[4], kernel_size=1, stride=1, padding=0),
            ConvBNReLU(cfg[4], cfg[5], kernel_size=1, stride=1, padding=0),
            nn.MaxPool2d(kernel_size=3, stride=2, padding=1),
            ConvBNReLU(cfg[5], cfg[6], kernel_size=3, stride=1, padding=1),
            ConvBNReLU(cfg[6], cfg[7], kernel_size=1, stride=1, padding=0),
            ConvBNReLU(cfg[7], 10, kernel_size=1, stride=1, padding=0),
            nn.AvgPool2d(kernel_size=8, stride=1, padding=0),
        )

    def forward(self, x):
        x = self.model(x)
        x = x.view(x.size(0), -1)
        return x



import copy

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Function


# ********************* quantizers（量化器，量化） *********************
# 取整(ste)
class Round(Function):
    @staticmethod
    def forward(self, input):
        sign = torch.sign(input)
        output = sign * torch.floor(torch.abs(input) + 0.5)
        return output

    @staticmethod
    def backward(self, grad_output):
        grad_input = grad_output.clone()
        return grad_input


# A(特征)量化
class ActivationQuantizer(nn.Module):
    def __init__(self, a_bits):
        super(ActivationQuantizer, self).__init__()
        self.a_bits = a_bits

    # 取整(ste)
    def round(self, input):
        output = Round.apply(input)
        return output

    # 量化/反量化
    def forward(self, input):
        if self.a_bits == 32:
            output = input
        elif self.a_bits == 1:
            print("！Binary quantization is not supported ！")
            assert self.a_bits != 1
        else:
            output = torch.clamp(input * 0.1, 0, 1)  # 特征A截断前先进行缩放（* 0.1），以减小截断误差
            scale = 1 / float(2 ** self.a_bits - 1)  # scale
            output = self.round(output / scale) * scale  # 量化/反量化
        return output


# W(权重)量化
class WeightQuantizer(nn.Module):
    def __init__(self, w_bits):
        super(WeightQuantizer, self).__init__()
        self.w_bits = w_bits

    # 取整(ste)
    def round(self, input):
        output = Round.apply(input)
        return output

    # 量化/反量化
    def forward(self, input):
        if self.w_bits == 32:
            output = input
        elif self.w_bits == 1:
            print("！Binary quantization is not supported ！")
            assert self.w_bits != 1
        else:
            output = torch.tanh(input)
            output = output / 2 / torch.max(torch.abs(output)) + 0.5  # 归一化-[0,1]
            scale = 1 / float(2 ** self.w_bits - 1)  # scale
            output = self.round(output / scale) * scale  # 量化/反量化
            output = 2 * output - 1
        return output


class QuantConv2d(nn.Conv2d):
    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        stride=1,
        padding=0,
        dilation=1,
        groups=1,
        bias=True,
        padding_mode="zeros",
        a_bits=8,
        w_bits=8,
        quant_inference=False,
    ):
        super(QuantConv2d, self).__init__(
            in_channels,
            out_channels,
            kernel_size,
            stride,
            padding,
            dilation,
            groups,
            bias,
            padding_mode,
        )
        self.quant_inference = quant_inference
        self.activation_quantizer = ActivationQuantizer(a_bits=a_bits)
        self.weight_quantizer = WeightQuantizer(w_bits=w_bits)

    def forward(self, input):
        quant_input = self.activation_quantizer(input)
        if not self.quant_inference:
            quant_weight = self.weight_quantizer(self.weight)
        else:
            quant_weight = self.weight
        output = F.conv2d(
            quant_input,
            quant_weight,
            self.bias,
            self.stride,
            self.padding,
            self.dilation,
            self.groups,
        )
        return output


class QuantConvTranspose2d(nn.ConvTranspose2d):
    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        stride=1,
        padding=0,
        output_padding=0,
        dilation=1,
        groups=1,
        bias=True,
        padding_mode="zeros",
        a_bits=8,
        w_bits=8,
        quant_inference=False,
    ):
        super(QuantConvTranspose2d, self).__init__(
            in_channels,
            out_channels,
            kernel_size,
            stride,
            padding,
            output_padding,
            dilation,
            groups,
            bias,
            padding_mode,
        )
        self.quant_inference = quant_inference
        self.activation_quantizer = ActivationQuantizer(a_bits=a_bits)
        self.weight_quantizer = WeightQuantizer(w_bits=w_bits)

    def forward(self, input):
        quant_input = self.activation_quantizer(input)
        if not self.quant_inference:
            quant_weight = self.weight_quantizer(self.weight)
        else:
            quant_weight = self.weight
        output = F.conv_transpose2d(
            quant_input,
            quant_weight,
            self.bias,
            self.stride,
            self.padding,
            self.output_padding,
            self.groups,
            self.dilation,
        )
        return output


class QuantLinear(nn.Linear):
    def __init__(
        self,
        in_features,
        out_features,
        bias=True,
        a_bits=8,
        w_bits=8,
        quant_inference=False,
    ):
        super(QuantLinear, self).__init__(in_features, out_features, bias)
        self.quant_inference = quant_inference
        self.activation_quantizer = ActivationQuantizer(a_bits=a_bits)
        self.weight_quantizer = WeightQuantizer(w_bits=w_bits)

    def forward(self, input):
        quant_input = self.activation_quantizer(input)
        if not self.quant_inference:
            quant_weight = self.weight_quantizer(self.weight)
        else:
            quant_weight = self.weight
        output = F.linear(quant_input, quant_weight, self.bias)
        return output


def add_quant_op(module, layer_counter, a_bits=8, w_bits=8, quant_inference=False):
    for name, child in module.named_children():
        if isinstance(child, nn.Conv2d):
            layer_counter[0] += 1
            if layer_counter[0] > 1:
                if child.bias is not None:
                    quant_conv = QuantConv2d(
                        child.in_channels,
                        child.out_channels,
                        child.kernel_size,
                        stride=child.stride,
                        padding=child.padding,
                        dilation=child.dilation,
                        groups=child.groups,
                        bias=True,
                        padding_mode=child.padding_mode,
                        a_bits=a_bits,
                        w_bits=w_bits,
                        quant_inference=quant_inference,
                    )
                    quant_conv.bias.data = child.bias
                else:
                    quant_conv = QuantConv2d(
                        child.in_channels,
                        child.out_channels,
                        child.kernel_size,
                        stride=child.stride,
                        padding=child.padding,
                        dilation=child.dilation,
                        groups=child.groups,
                        bias=False,
                        padding_mode=child.padding_mode,
                        a_bits=a_bits,
                        w_bits=w_bits,
                        quant_inference=quant_inference,
                    )
                quant_conv.weight.data = child.weight
                module._modules[name] = quant_conv
        elif isinstance(child, nn.ConvTranspose2d):
            layer_counter[0] += 1
            if layer_counter[0] > 1:
                if child.bias is not None:
                    quant_conv_transpose = QuantConvTranspose2d(
                        child.in_channels,
                        child.out_channels,
                        child.kernel_size,
                        stride=child.stride,
                        padding=child.padding,
                        output_padding=child.output_padding,
                        dilation=child.dilation,
                        groups=child.groups,
                        bias=True,
                        padding_mode=child.padding_mode,
                        a_bits=a_bits,
                        w_bits=w_bits,
                        quant_inference=quant_inference,
                    )
                    quant_conv_transpose.bias.data = child.bias
                else:
                    quant_conv_transpose = QuantConvTranspose2d(
                        child.in_channels,
                        child.out_channels,
                        child.kernel_size,
                        stride=child.stride,
                        padding=child.padding,
                        output_padding=child.output_padding,
                        dilation=child.dilation,
                        groups=child.groups,
                        bias=False,
                        padding_mode=child.padding_mode,
                        a_bits=a_bits,
                        w_bits=w_bits,
                        quant_inference=quant_inference,
                    )
                quant_conv_transpose.weight.data = child.weight
                module._modules[name] = quant_conv_transpose
        elif isinstance(child, nn.Linear):
            layer_counter[0] += 1
            if layer_counter[0] > 1:
                if child.bias is not None:
                    quant_linear = QuantLinear(
                        child.in_features,
                        child.out_features,
                        bias=True,
                        a_bits=a_bits,
                        w_bits=w_bits,
                        quant_inference=quant_inference,
                    )
                    quant_linear.bias.data = child.bias
                else:
                    quant_linear = QuantLinear(
                        child.in_features,
                        child.out_features,
                        bias=False,
                        a_bits=a_bits,
                        w_bits=w_bits,
                        quant_inference=quant_inference,
                    )
                quant_linear.weight.data = child.weight
                module._modules[name] = quant_linear
        else:
            add_quant_op(
                child,
                layer_counter,
                a_bits=a_bits,
                w_bits=w_bits,
                quant_inference=quant_inference,
            )


def prepare(model, inplace=False, a_bits=8, w_bits=8, quant_inference=False):
    if not inplace:
        model = copy.deepcopy(model)
    layer_counter = [0]
    add_quant_op(
        model,
        layer_counter,
        a_bits=a_bits,
        w_bits=w_bits,
        quant_inference=quant_inference,
    )
    return model



import sys

sys.path.append("../..")
import os
import argparse
import numpy as np
import torch
import torch.nn as nn
from torch.autograd import Variable
from torchvision import datasets, transforms
from models import nin_gc


parser = argparse.ArgumentParser()
parser.add_argument("--data", action="store", default="../../data", help="dataset path")
parser.add_argument("--cpu", action="store_true", help="disables CUDA training")
parser.add_argument("--percent", type=float, default=0.4, help="nin_gc:0.4")
parser.add_argument("--layers", type=int, default=9, help="layers (default: 9)")
parser.add_argument(
    "--model",
    default="models_save/nin_gc_preprune.pth",
    type=str,
    metavar="PATH",
    help="path to raw trained model (default: none)",
)
args = parser.parse_args()
layers = args.layers
print(args)

model = nin_gc.Net()
if args.model:
    if os.path.isfile(args.model):
        print("=> loading checkpoint '{}'".format(args.model))
        model.load_state_dict(torch.load(args.model)["state_dict"])
    else:
        print("=> no checkpoint found at '{}'".format(args.resume))
print("旧模型: ", model)
total = 0
i = 0
for m in model.modules():
    if isinstance(m, nn.BatchNorm2d):
        if i < layers - 1:
            i += 1
            total += m.weight.data.shape[0]

bn = torch.zeros(total)
index = 0
i = 0
for m in model.modules():
    if isinstance(m, nn.BatchNorm2d):
        if i < layers - 1:
            i += 1
            size = m.weight.data.shape[0]
            bn[index : (index + size)] = m.weight.data.abs().clone()
            index += size
y, j = torch.sort(bn)
thre_index = int(total * args.percent)
if thre_index == total:
    thre_index = total - 1
thre_0 = y[thre_index]

# ****************获取针对分组卷积剪枝每层的基数（base_bumber）********************
nums = []
channels = []
groups = [1]
prune_base_num = []
j = 0
for k, m in enumerate(model.modules()):
    if isinstance(m, nn.Conv2d):
        s_0 = m.weight.data.data.size()[0]
        s_1 = m.weight.data.data.size()[1]
        nums.append(s_0)
        channels.append(s_1)
while j < len(nums) - 1:
    groups.append(int(nums[j] / channels[j + 1]))
    j += 1
# print(groups)
j = 0
while j < len(groups) - 1:
    for i in range(1, (groups[j] * groups[j + 1]) + 1):
        if i % groups[j] == 0 and i % groups[j + 1] == 0:
            prune_base_num.append(i)
            break
    j += 1
# print(prune_base_num)

# ********************************预剪枝*********************************
pruned = 0
cfg_0 = []
cfg = []
cfg_mask = []
i = 0
for k, m in enumerate(model.modules()):
    if isinstance(m, nn.BatchNorm2d):
        if i < layers - 1:
            i += 1

            weight_copy = m.weight.data.clone()
            mask = weight_copy.abs().gt(thre_0).float()
            remain_channels = torch.sum(mask)

            if remain_channels == 0:
                print("\r\n!please turn down the prune_ratio!\r\n")
                remain_channels = 1
                mask[int(torch.argmax(weight_copy))] = 1

            # ******************分组卷积剪枝******************
            base_number = prune_base_num[i - 1]
            v = 0
            n = 1
            if remain_channels % base_number != 0:
                if remain_channels > base_number:
                    while v < remain_channels:
                        n += 1
                        v = base_number * n
                    if remain_channels - (v - base_number) < v - remain_channels:
                        remain_channels = v - base_number
                    else:
                        remain_channels = v
                    if remain_channels > m.weight.data.size()[0]:
                        remain_channels = m.weight.data.size()[0]
                    remain_channels = torch.tensor(remain_channels)

                    y, j = torch.sort(weight_copy.abs())
                    thre_1 = y[-remain_channels]
                    mask = weight_copy.abs().ge(thre_1).float()
            pruned = pruned + mask.shape[0] - torch.sum(mask)
            m.weight.data.mul_(mask)
            m.bias.data.mul_(mask)
            cfg_0.append(mask.shape[0])
            cfg.append(int(remain_channels))
            cfg_mask.append(mask.clone())
            print(
                "layer_index: {:d} \t total_channel: {:d} \t remaining_channel: {:d} \t pruned_ratio: {:f}".format(
                    k,
                    mask.shape[0],
                    int(torch.sum(mask)),
                    (mask.shape[0] - torch.sum(mask)) / mask.shape[0],
                )
            )
pruned_ratio = float(pruned / total)
print("\r\n!预剪枝完成!")
print("total_pruned_ratio: ", pruned_ratio)

# ********************************预剪枝后model测试*********************************
def test():
    test_loader = torch.utils.data.DataLoader(
        datasets.CIFAR10(
            root=args.data,
            train=False,
            transform=transforms.Compose(
                [
                    transforms.ToTensor(),
                    transforms.Normalize(
                        (0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010)
                    ),
                ]
            ),
        ),
        batch_size=64,
        shuffle=False,
        num_workers=1,
    )
    model.eval()
    correct = 0

    for data, target in test_loader:
        if not args.cpu:
            data, target = data.cuda(), target.cuda()
        data, target = Variable(data), Variable(target)
        output = model(data)
        pred = output.data.max(1, keepdim=True)[1]
        correct += pred.eq(target.data.view_as(pred)).cpu().sum()
    acc = 100.0 * float(correct) / len(test_loader.dataset)
    print("Accuracy: {:.2f}%\n".format(acc))
    return


print("************预剪枝模型测试************")
if not args.cpu:
    model.cuda()
test()

# *****************************剪枝后model结构********************************
newmodel = nin_gc.Net(cfg)
print("新模型: ", newmodel)

# *****************************剪枝前后model对比********************************
print("************旧模型结构************")
print(cfg_0)
print("************新模型结构************")
print(cfg, "\r\n")



import torch.nn as nn


def channel_shuffle(x, groups):
    """shuffle channels of a 4-D Tensor"""
    batch_size, channels, height, width = x.size()
    assert channels % groups == 0
    channels_per_group = channels // groups
    # split into groups
    x = x.view(batch_size, groups, channels_per_group, height, width)
    # transpose 1, 2 axis
    x = x.transpose(1, 2).contiguous()
    # reshape into orignal
    x = x.view(batch_size, channels, height, width)
    return x


class ConvBNReLU(nn.Module):
    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        stride=1,
        padding=0,
        dilation=1,
        groups=1,
        bias=True,
        padding_mode="zeros",
        eps=1e-5,
        momentum=0.1,
        channel_shuffle=0,
        shuffle_groups=1,
    ):
        super(ConvBNReLU, self).__init__()
        self.channel_shuffle_flag = channel_shuffle
        self.shuffle_groups = shuffle_groups

        self.conv = nn.Conv2d(
            in_channels,
            out_channels,
            kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            groups=groups,
            bias=bias,
            padding_mode=padding_mode,
        )
        self.bn = nn.BatchNorm2d(out_channels, eps=eps, momentum=momentum)
        self.relu = nn.ReLU(inplace=True)

    def forward(self, x):
        if self.channel_shuffle_flag:
            x = channel_shuffle(x, groups=self.shuffle_groups)
        x = self.conv(x)
        x = self.bn(x)
        x = self.relu(x)
        return x


class Net(nn.Module):
    def __init__(self, cfg=None):
        super(Net, self).__init__()
        if cfg is None:
            cfg = [256, 256, 256, 512, 512, 512, 1024, 1024]
        self.model = nn.Sequential(
            ConvBNReLU(3, cfg[0], kernel_size=5, stride=1, padding=2),
            ConvBNReLU(
                cfg[0],
                cfg[1],
                kernel_size=1,
                stride=1,
                padding=0,
                groups=2,
                channel_shuffle=0,
            ),
            ConvBNReLU(
                cfg[1],
                cfg[2],
                kernel_size=1,
                stride=1,
                padding=0,
                groups=2,
                channel_shuffle=1,
                shuffle_groups=2,
            ),
            nn.MaxPool2d(kernel_size=2, stride=2, padding=0),
            ConvBNReLU(
                cfg[2],
                cfg[3],
                kernel_size=3,
                stride=1,
                padding=1,
                groups=16,
                channel_shuffle=1,
                shuffle_groups=2,
            ),
            ConvBNReLU(
                cfg[3],
                cfg[4],
                kernel_size=1,
                stride=1,
                padding=0,
                groups=4,
                channel_shuffle=1,
                shuffle_groups=16,
            ),
            ConvBNReLU(
                cfg[4],
                cfg[5],
                kernel_size=1,
                stride=1,
                padding=0,
                groups=4,
                channel_shuffle=1,
                shuffle_groups=4,
            ),
            nn.MaxPool2d(kernel_size=2, stride=2, padding=0),
            ConvBNReLU(
                cfg[5],
                cfg[6],
                kernel_size=3,
                stride=1,
                padding=1,
                groups=32,
                channel_shuffle=1,
                shuffle_groups=4,
            ),
            ConvBNReLU(
                cfg[6],
                cfg[7],
                kernel_size=1,
                stride=1,
                padding=0,
                groups=8,
                channel_shuffle=1,
                shuffle_groups=32,
            ),
            ConvBNReLU(cfg[7], 10, kernel_size=1, stride=1, padding=0),
            nn.AvgPool2d(kernel_size=8, stride=1, padding=0),
        )

    def forward(self, x):
        x = self.model(x)
        x = x.view(x.size(0), -1)
        return x



import torch
import torch.nn as nn


class Add(nn.Module):
    def __init__(self):
        super(Add, self).__init__()

    def forward(self, res, shortcut):
        output = res + shortcut
        return output


# *** temp_dev ***
"""
class Concat(nn.Module):
    def __init__(self, dim=1):
        super(Concat, self).__init__()
        self.dim = dim

    def forward(self, res, shortcut):
        output = torch.cat([shortcut, res], dim=self.dim)
        return output
"""



# test_trt(temp_demo)

import numpy as np
import torch
import torch.nn as nn

# tensorrt-lib
import util_trt
from calibrator import SegBatchStream


# trt test function
def test_trt(segmentation_module_trt, loader, gpu):
    infer_time_sum = 0.0
    warmup_flag = 0
    warmup_nums = 5
    # pbar = tqdm(total=len(loader))

    for batch_data in loader:
        warmup_flag += 1
        # process data
        batch_data = batch_data[0]
        segSize = (batch_data["img_ori"].shape[0], batch_data["img_ori"].shape[1])
        img_resized_list = batch_data["img_data"]
        resized_nums = len(img_resized_list)

        with torch.no_grad():
            scores = torch.zeros(1, cfg.DATASET.num_class, segSize[0], segSize[1])
            scores = async_copy_to(scores, gpu)

            for img in img_resized_list:
                feed_dict = batch_data.copy()
                feed_dict["img_data"] = img
                del feed_dict["img_ori"]
                del feed_dict["info"]
                feed_dict = async_copy_to(feed_dict, gpu)

                # forward pass
                pred_tmp, infer_time = segmentation_module_trt(
                    feed_dict, segSize=segSize, shape_of_input=img.shape
                )
                if warmup_flag > warmup_nums:
                    infer_time_sum = infer_time_sum + infer_time
                scores = scores + pred_tmp.cuda() / len(cfg.DATASET.imgSizes)

            _, pred = torch.max(scores, dim=1)
            pred = as_numpy(pred.squeeze(0).cpu())

            # import pdb
            # pdb.set_trace()
            # cv2.imwrite('sl.jpg', (pred*255).astype(np.uint8))

        # Judge whether the image is saliency
        if (
            np.mean(scores.squeeze(0)[1, :, :].squeeze(0).cpu().numpy()) < 0.01
            or np.mean(pred) < 0.001
        ):
            pred = as_numpy(torch.zeros(segSize[0], segSize[1]))
            scores = torch.zeros(1, cfg.DATASET.num_class, segSize[0], segSize[1])
            scores = async_copy_to(scores, gpu)

        # Connection Region Judge
        saliecy_map = scores.squeeze(0)[1, :, :].squeeze(0).cpu().numpy()
        saliecy_map *= 255
        ret, thresh = cv2.threshold(saliecy_map.astype(np.uint8), 127, 255, 0)
        contours, hierarchy = cv2.findContours(
            thresh, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE
        )
        # print(len(contours), batch_data['info'])
        # if len(contours) > 5
        valid_region = 0
        max_region = 0
        for idx, contour in enumerate(contours):
            x1, y1 = np.min(contour[:, :, 0]), np.min(contour[:, :, 1])
            x2, y2 = np.max(contour[:, :, 0]), np.max(contour[:, :, 1])
            if hierarchy[0][idx][-1] != -1:  # Children Region
                continue
            valid_region += 1
            area = (x2 - x1) * (y2 - y1)
            if area < 300:  # Area Filter
                pred[y1:y2, x1:x2] = 0
                continue
            if area > max_region:
                max_region = area

        # print(valid_region, len(contours), batch_data['info'])

        # 黑色背景
        visualize_result_black_trt(
            (batch_data["img_ori"], batch_data["info"]),
            pred,
            cfg,
            scores.squeeze(0)[1, :, :].squeeze(0).cpu().numpy(),
        )
        # pbar.update(1)
    fps = (len(loader) - warmup_nums) * resized_nums / infer_time_sum
    return fps


def main(cfg, gpu):
    # Dataset and Loader
    dataset_test = TestDataset(cfg.list_test, cfg.DATASET)
    loader_test = torch.utils.data.DataLoader(
        dataset_test,
        batch_size=cfg.TEST.batch_size,
        shuffle=False,
        collate_fn=user_scattered_collate,
        num_workers=5,
        drop_last=True,
    )

    # model
    net_resnet = build_resnet_upsample(
        arch=cfg.MODEL.arch_encoder.lower(),
        arch_de=cfg.MODEL.arch_decoder.lower(),
        fc_dim=cfg.MODEL.fc_dim,
        fc_dim_de=cfg.MODEL.fc_dim,
        num_class_de=cfg.DATASET.num_class,
        use_softmax=True,
        deep_sup_scale=None,
        segSize=None,
    )
    net_resnet.load_state_dict(
        torch.load(cfg.MODEL.weights_seg, map_location=lambda storage, loc: storage),
        strict=True,
    )

    # torch2onnx
    input_name = ["input"]
    output_name = ["output"]
    onnx_model_fixed = "models_save/model_seg_fixed.onnx"
    onnx_model_dynamic = "models_save/model_seg_dynamic.onnx"
    batch_size = 1
    img_size_fixed = (3, 400, 400)
    img_size_dynamic = (3, 800, 800)
    dummy_input_fixed = torch.rand(batch_size, *img_size_fixed)
    dummy_input_dynamic = torch.rand(batch_size, *img_size_dynamic)
    dynamic_axes = {"input": {2: "height", 3: "width"}}
    print("\n*** torch to onnx begin ***")
    # fixed_onnx
    torch.onnx.export(
        net_resnet,
        dummy_input_fixed,
        onnx_model_fixed,
        verbose=True,
        input_names=input_name,
        output_names=output_name,
        opset_version=10,
    )
    # dynamic_onnx
    torch.onnx.export(
        net_resnet,
        dummy_input_dynamic,
        onnx_model_dynamic,
        verbose=True,
        input_names=input_name,
        output_names=output_name,
        opset_version=10,
        dynamic_axes=dynamic_axes,
    )
    print("*** torch to onnx completed ***\n")

    # onnx2trt
    fp16_mode = False
    int8_mode = True
    transform = None
    print("*** onnx to tensorrt begin ***")
    max_calibration_size = 80  # 校准集数量
    calibration_batche_size = 10  # 校准batch_size
    max_calibration_batches = max_calibration_size / calibration_batche_size
    # calibration
    calibration_stream = SegBatchStream(
        dataset_test,
        transform,
        calibration_batche_size,
        img_size_fixed,
        max_batches=max_calibration_batches,
    )
    engine_model_fixed = "models_save/model_seg_fixed.trt"
    engine_model_dynamic = "models_save/model_seg_dynamic.trt"
    calibration_table = "models_save/calibration_seg.cache"
    # fixed_engine,校准产生校准表
    engine_fixed = util_trt.get_engine(
        batch_size,
        onnx_model_fixed,
        engine_model_fixed,
        fp16_mode=fp16_mode,
        int8_mode=int8_mode,
        calibration_stream=calibration_stream,
        calibration_table_path=calibration_table,
        save_engine=True,
        dynamic=False,
    )
    assert engine_fixed, "Broken engine_fixed"
    print("*** engine_fixed completed ***\n")
    # dynamic_engine,加载fixed_engine生成的校准表,用于inference
    engine_dynamic = util_trt.get_engine(
        batch_size,
        onnx_model_dynamic,
        engine_model_dynamic,
        fp16_mode=fp16_mode,
        int8_mode=int8_mode,
        calibration_stream=calibration_stream,
        calibration_table_path=calibration_table,
        save_engine=True,
        dynamic=True,
    )
    assert engine_dynamic, "Broken engine_dynamic"
    print("*** engine_dynamic completed ***\n")
    print("*** onnx to tensorrt completed ***\n")
    # context and buffer
    context = engine_dynamic.create_execution_context()
    # choose an optimization profile
    context.active_optimization_profile = 0
    buffers = util_trt.allocate_buffers_v2(engine_dynamic, 1200, 1200)

    # trt test
    crit = nn.NLLLoss(ignore_index=-1)
    segmentation_module_trt = SegmentationModule_v2_trt(
        context, buffers, crit, use_softmax=True, binding_id=0
    )
    segmentation_module_trt.cuda()
    # result
    print("*** trt_model ***")
    print("test ing...")
    fps_trt = test_trt(segmentation_module_trt, loader_test, gpu)
    print("trt_model test done!")
    """
    print('trt-fps: ', fps_trt)
    """


if __name__ == "__main__":
    main(cfg, args.gpu)
